/**
 * filename: dcjdDist.cpp
 * author: Martin Bader
 * begin: 20.12.2007
 * last change: 22.03.2010
 *
 * Implementation of dcjdDist.h
 * 
 * Copyright (C) 2007  Martin Bader
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
  
#include "dcjdDist.h"

#include <climits>
#include <iostream>
#include <stdlib.h> 

#define DEBSEQ 0

namespace dcjdDist
{
using namespace std;

/***************************************************************************
 * public methods
 ***************************************************************************/
 
DcjdDist::DcjdDist()
{
    srand48(1);
    m_numMarkers = 0;
    m_permSize = 0;
    m_perm = NULL;         
    m_rddDegree = NULL; 
    m_rddEdges = NULL;
    m_compID = 0;
    m_tmpPerm = NULL;
    m_timestamp = NULL;
    m_numArrivingBacklinks = NULL;
    m_arrivingBacklinks = NULL;
    m_backlinkSource = NULL;
    m_backlinkTarget = NULL;
    m_timestamp2Node = NULL;
    m_bridgeStatus = NULL;
    m_bridges = NULL;
    m_bridgeCompID = NULL;
    m_maxChild = NULL;
    m_numSelfloops = NULL;
}         


DcjdDist::~DcjdDist()
{
}


void DcjdDist::sort(int f_numMarkers, int f_permSize, int* f_genome)
{
    int u, v;          // two nodes of an edge
    unsigned int opIndex;       // index of the next operation in m_goodMoves
    int numGoodMoves = 0;       // number of performed good moves
    int numMoves = 0;           // number of performed moves (overall weight)
    int lb;                     // the actual lower bound
    int* seqscore;              // score of a sequence
    int score;                  // score v 1.0
    Operation* op;              // operation of the sorting sequence -> push on stack
    int* perm;                  // permutation on the sorting sequence -> push in stack
    
    // set the permutation; this should be an augmented permutation
    m_numMarkers = f_numMarkers;
    m_permSize = f_permSize;
    m_perm = f_genome;
    if (m_permSize > m_numMarkers)
        m_tmpPerm = new int[50 * m_permSize];
    else        
        m_tmpPerm = new int[50 * m_numMarkers];
    // create the RDD  
    m_rddDegree = new int[m_numMarkers * 2 - 2];
    m_rddEdges = new int*[m_numMarkers * 2 - 2];
    if (m_permSize > m_numMarkers)
    {
        for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
            m_rddEdges[i] = new int[50 * m_permSize];       // make it big...
    }
    else
    {
        for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
            m_rddEdges[i] = new int[50 * m_numMarkers];       // make it big...
    }
    for (int i = 0; i < m_numMarkers - 1; i++)     // desire edges
    {
        m_rddEdges[2*i][0] = 2*i+1;
        m_rddEdges[2*i+1][0] = 2*i;
        m_rddDegree[2*i] = 1;
        m_rddDegree[2*i+1] = 1;
    }
    for (int i = 0; i < m_permSize - 1; i++)       // reality edges
    {
        u = (m_perm[i] >= 0? 2 * m_perm[i] : 2 * -m_perm[i] - 1);
        v = (m_perm[i+1] < 0? 2 * -m_perm[i+1] : 2 * m_perm[i+1] - 1);
        m_rddEdges[u][m_rddDegree[u]++] = v;
        m_rddEdges[v][m_rddDegree[v]++] = u;
    }
    // init selfloops
    m_numSelfloops = new int[m_numMarkers * 2 - 2];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        m_numSelfloops[i] = 0;
        for (int j = 0; j < m_rddDegree[i]; j++)
        {
            if (m_rddEdges[i][j] == i)
                m_numSelfloops[i]++;
        }
        m_numSelfloops[i] /= 2;
    } 
    cout << "lower bound: " << getLowerBound() << endl;
    while (true)
    {
        // find possible moves
        lb = getLowerBound();
        initBridges();
        findBridges();
        listGoodMoves();
        for (unsigned int i = 0; i < m_seqscores.size(); i++)
            delete[] m_seqscores[i];
        m_seqscores.clear();
        for (unsigned int i = 0; i < m_goodMoves.size(); i++)
        {
            if (m_goodMoves[i]->m_type == Operation::OPERATION_DCJCIRC)
            {
                getAllFollowUps(*(m_goodMoves[i]));
                m_scores.push_back(-1);
                seqscore = new int[4];
                seqscore[0] = -1;
                m_seqscores.push_back(seqscore);
            }
            else
            {
                seqscore = new int[4];
                score = getScore(*(m_goodMoves[i]), lb, seqscore);
                m_scores.push_back(score);
                m_seqscores.push_back(seqscore);
            }    
        }
        opIndex = 0;
        for (unsigned int i = 1; i < m_goodMoves.size(); i++)
        {
            if (compSeqScores(m_seqscores[i], m_seqscores[opIndex]) == 1)
                opIndex = i;
        }

        if (!((opIndex >= m_goodMoves.size()) || (m_scores[opIndex] <= 0)))
        {
            if ((m_goodMoves[opIndex]->m_type == Operation::OPERATION_BLOCKINTERCHANGE)
              && (m_seqscores[opIndex][0] == 60))
                numGoodMoves += 2;
            else
                numGoodMoves++;
        }
        opIndex = 0;
        for (unsigned int i = 1; i < m_goodMoves.size(); i++)
        {
            if (compSeqScores(m_seqscores[i], m_seqscores[opIndex]) == 1)
                opIndex = i;
        }

        if ((opIndex < m_goodMoves.size()) && (m_scores[opIndex] > 0))
        {
            // push operation for later output
            op = new Operation;
            op->m_type = m_goodMoves[opIndex]->m_type;
            op->m_edges[0] = m_goodMoves[opIndex]->m_edges[0];
            op->m_edges[1] = m_goodMoves[opIndex]->m_edges[1];
            op->m_edges[2] = m_goodMoves[opIndex]->m_edges[2];
            op->m_edges[3] = m_goodMoves[opIndex]->m_edges[3];
            for (unsigned int i = 0; i < m_goodMoves[opIndex]->m_sequence.size(); i++)
                op->m_sequence.push_back(m_goodMoves[opIndex]->m_sequence[i]);
            m_sortingSequence.push(op);
            m_sortingSizes.push(m_permSize);
            perm = new int[m_permSize];
            for (int i = 0; i < m_permSize; i++)
                perm[i] = m_perm[i];
            m_sortingPerms.push(perm);
            // perform the operation
            performOperation(*(m_goodMoves[opIndex]));
            // assert everything was ok
            if (!checkRDD())
                return;
            if (m_goodMoves[opIndex]->m_type == Operation::OPERATION_BLOCKINTERCHANGE)
                numMoves++;
            numMoves++;                        
        }
        else        
        {
            for (unsigned int i = 0; i < m_sequences.size(); i++)
            {
                for (unsigned int j = 0; j < m_sequences[i].size(); j++)
                {
                    m_sequences[i][j]->m_sequence.clear();
                    delete m_sequences[i][j];
                }
            } 
            m_sequences.clear();
            for (unsigned int i = 0; i < m_seqscores.size(); i++)
                delete[] m_seqscores[i];
            m_seqscores.clear();
            findReverseDuplications();
            findAdjacencySequences();
            findType2Sequences();
            findType3Sequences();
            findTranspositionDuplications();
            findSpecialSequences();
            for (unsigned int i = 0; i < m_sequences.size(); i++)
            {
                seqscore = new int[4];
                getSequenceScore(m_sequences[i], lb, seqscore);
                m_seqscores.push_back(seqscore);
            }
            opIndex = 0;
            for (unsigned int i = 1; i < m_sequences.size(); i++)
            {
                if (compSeqScores(m_seqscores[i], m_seqscores[opIndex]) == 1)
                    opIndex = i;
            }
            if (opIndex < m_sequences.size()) 
            {
                for (unsigned int i = 0; i < m_sequences[opIndex].size(); i++)
                {
                    // push operation for later output
                    op = new Operation;
                    op->m_type = m_sequences[opIndex][i]->m_type;
                    op->m_edges[0] = m_sequences[opIndex][i]->m_edges[0];
                    op->m_edges[1] = m_sequences[opIndex][i]->m_edges[1];
                    op->m_edges[2] = m_sequences[opIndex][i]->m_edges[2];
                    op->m_edges[3] = m_sequences[opIndex][i]->m_edges[3];
                    for (unsigned int j = 0; j < m_sequences[opIndex][i]->m_sequence.size(); j++)
                        op->m_sequence.push_back(m_sequences[opIndex][i]->m_sequence[j]);
                    m_sortingSequence.push(op);
                    m_sortingSizes.push(m_permSize);
                    perm = new int[m_permSize];
                    for (int j = 0; j < m_permSize; j++)
                        perm[j] = m_perm[j];
                    m_sortingPerms.push(perm);
                    // perform the operation
                    performOperation(*(m_sequences[opIndex][i]));
                    // check if everything was ok
                    if (!checkRDD())
                        return;
                    if (m_sequences[opIndex][i]->m_type == Operation::OPERATION_BLOCKINTERCHANGE)
                        numMoves += 2;
                    else
                        numMoves++;    
                }
            }
            else
                break;
        }
    }
    // print the sorting
    for (int i = 1; i < m_permSize - 1; i++)
        cout << m_perm[i] << " ";
    cout << endl;
    while (!m_sortingSequence.empty())
    {
        op = m_sortingSequence.top();
        if (op->m_type == Operation::OPERATION_DELETION)
            cout << "Del(" << op->m_edges[0] << "," << op->m_edges[0] + op->m_sequence.size() << ")\n";   
        else
            printOperation(*op);
        perm = m_sortingPerms.top();
        for (int i = 1; i < m_sortingSizes.top() - 1; i++)
            cout << perm[i] << " ";
        cout << endl;
        delete op;
        delete[] perm;
        m_sortingSequence.pop();
        m_sortingSizes.pop();
        m_sortingPerms.pop();
    }
    cout << "distance: " << numMoves << endl;        
    // clean up
    delete[] m_rddDegree; m_rddDegree = NULL;
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        delete[] m_rddEdges[i];
    delete[] m_rddEdges; m_rddEdges = NULL;
    clearBridges();
    delete[] m_numSelfloops; m_numSelfloops = NULL;
    for (unsigned int i = 0; i < m_goodMoves.size(); i++)
        delete m_goodMoves[i];
    m_goodMoves.clear();
    for (unsigned int i = 0; i < m_sequences.size(); i++)
    {
        for (unsigned int j = 0; j < m_sequences[i].size(); j++)
        {
            m_sequences[i][j]->m_sequence.clear();
            delete m_sequences[i][j];
        }
    } 
    m_sequences.clear();
    for (unsigned int i = 0; i < m_seqscores.size(); i++)
        delete[] m_seqscores[i];
    m_seqscores.clear();
    delete[] m_tmpPerm; m_tmpPerm = NULL;
}  


/***************************************************************************
 * private methods
 ***************************************************************************/

void DcjdDist::initBridges()
{
    m_timestamp = new int[m_numMarkers * 2 - 2];
    m_numArrivingBacklinks = new int[m_numMarkers * 2 - 2];
    m_arrivingBacklinks = new int*[m_numMarkers * 2 - 2];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        m_arrivingBacklinks[i] = new int[m_numMarkers * 2 - 2];
    m_numOpenBacklinks = new int[m_numMarkers * 2 - 2];
    m_backlinkSource = new int[m_numMarkers * 2 - 2];
    m_backlinkTarget = new int[m_numMarkers * 2 - 2];
    m_longestBacklink = new int[m_numMarkers * 2 - 2]; 
    m_intervalStacks = new stack<int>*[m_numMarkers * 2 - 2];
    m_intervalStores = new stack<int>*[m_numMarkers * 2 - 2];
    m_openInterval = new bool[m_numMarkers * 2 - 2];
    m_timestamp2Node = new int[m_numMarkers * 2 - 2];
    m_bridgeStatus = new int[m_numMarkers * 2 - 2];
    m_bridges = new pair<int, int>*[m_numMarkers * 2 - 2];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        m_bridges[i] = new pair<int, int>[m_numMarkers * 2 - 2];
    m_bridgeCompID = new int[m_numMarkers * 2 - 2];
    m_maxChild = new int[m_numMarkers * 2 - 2];
}


void DcjdDist::findBridges()
{

    // init
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        m_timestamp[i] = -1;
    m_compID = 0;        
    // depth first search
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        if (m_timestamp[i] != -1)
            continue;
        findBridgesExpand(i, 0, -1);
        // clean up
        delete m_intervalStacks[i];
        delete m_intervalStores[i];
        m_compID++;           // start new component
    }
    // adjust bridge markers    
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        if (m_bridgeStatus[i] == 1)           // no bridge
            m_bridgeStatus[i] = 0;
        else if (m_bridgeStatus[i] == -1)     // 1-bridge
            m_bridgeStatus[i] = 1;
    }
}


int DcjdDist::findBridgesExpand(int f_node, int f_time, int f_parent)
{
    bool foundParent = false;                             // don't expand the same edge we came from
    int child;                                            // index of a child node
    int ts;                                               // timestamp of a child node
    int pushSource = -1;                                  // source of an interval to push on the stack

    // init the node    
    m_timestamp[f_node] = f_time;
    m_numArrivingBacklinks[f_node] = 0;
    m_numOpenBacklinks[f_node] = 0;
    m_backlinkSource[f_node] = f_node;
    m_backlinkTarget[f_node] = f_node;
    m_longestBacklink[f_node] = f_node;
    m_intervalStacks[f_node] = new stack<int>();
    m_intervalStores[f_node] = new stack<int>();    
    m_timestamp2Node[f_time] = f_node;
    m_bridgeStatus[f_node] = 0;
    m_bridgeCompID[f_node] = m_compID;
    m_maxChild[f_node] = f_time;
    f_time++;
    // check all neighbors
    for (int i = 0; i < m_rddDegree[f_node]; i++) 
    {
        if (!foundParent && (m_rddEdges[f_node][i] == f_parent))   // we came this edge
        {
            foundParent = true;
            continue;
        }
        child = m_rddEdges[f_node][i];
        ts = m_timestamp[child];
        if ((ts != -1) && (ts < m_timestamp[f_node]))              // this is a backlink
        {
            m_arrivingBacklinks[child][m_numArrivingBacklinks[child]] = f_node; 
            m_numArrivingBacklinks[child]++;
            if (ts <= m_timestamp[m_backlinkTarget[f_node]])
            {
                m_backlinkTarget[f_node] = child;
                m_backlinkSource[f_node] = f_node;
                pushSource = f_node;
            }
            if (ts < m_timestamp[m_longestBacklink[f_node]])
                m_longestBacklink[f_node] = child;
            m_numOpenBacklinks[f_node]++;
        }
        else if (ts == -1)                        // true child - expand
        {
            f_time = findBridgesExpand(child, f_time, f_node);
            m_maxChild[f_node] = m_maxChild[child];
            m_numOpenBacklinks[f_node] += m_numOpenBacklinks[child];
            if (m_timestamp[m_backlinkTarget[child]] <= m_timestamp[m_backlinkTarget[f_node]])  // child has longer backlink
            {
                delete m_intervalStacks[f_node]; m_intervalStacks[f_node] = NULL;
                delete m_intervalStores[f_node]; m_intervalStores[f_node] = NULL;
                m_intervalStacks[f_node] = m_intervalStacks[child];
                m_intervalStores[f_node] = m_intervalStores[child];
                if (m_backlinkTarget[f_node] != f_node)
                    pushSource = m_backlinkSource[f_node];
                m_backlinkSource[f_node] = m_backlinkSource[child];
                m_backlinkTarget[f_node] = m_backlinkTarget[child];
            }
            else
            {
                delete m_intervalStacks[child]; m_intervalStacks[child] = NULL;
                delete m_intervalStores[child]; m_intervalStores[child] = NULL;
                if (m_timestamp[m_backlinkTarget[child]] < m_timestamp[m_longestBacklink[f_node]])
                    pushSource = m_backlinkSource[child]; 
            }
        }
        // else another link to a child node; already used as backlink in opposite direction
    }
    // handle incoming backlinks
    for (int i = 0; i < m_numArrivingBacklinks[f_node]; i++)
        m_openInterval[m_arrivingBacklinks[f_node][i]] = false;
    m_numOpenBacklinks[f_node] -= m_numArrivingBacklinks[f_node];
    // remove closed intervals
    while (!m_intervalStacks[f_node]->empty() && !m_openInterval[m_intervalStacks[f_node]->top()])
    {
        if (m_timestamp[f_node] <= m_timestamp[m_longestBacklink[m_intervalStacks[f_node]->top()]])
        {
            m_intervalStacks[f_node]->pop();
            m_intervalStores[f_node]->pop();
        }
        else
        {
            m_openInterval[m_intervalStacks[f_node]->top()] = true;
            m_intervalStores[f_node]->pop();
            m_intervalStores[f_node]->push(-1);
        }
    }
    // check if we must open a new interval
    if (pushSource != -1)
    {
        m_intervalStacks[f_node]->push(pushSource);
        m_intervalStores[f_node]->push(-1);
        m_openInterval[pushSource] = true;
    }
    else if (m_intervalStores[f_node]->empty())
    {
        m_intervalStacks[f_node]->push(f_node);
        m_intervalStores[f_node]->push(-1);
        m_openInterval[f_node] = true;
    }
    // report if edge to parent is in a bridge
    if (m_numOpenBacklinks[f_node] == 0)   
    {
        if (f_parent != -1)                 // this is a 1-bridge
        {
            m_bridgeStatus[f_node] = -1;
            m_bridges[f_node][0].first = f_node;
            m_bridges[f_node][0].second = f_parent;
        }
    }
    else                                    // this might be a 2-bridge
    {
        if (m_intervalStores[f_node]->top() == -1)       // first report for this interval
        {
            if (m_numOpenBacklinks[f_node] == 1)        // 2-bridge with backlink -> report backlink
            {
                m_bridges[f_node][0].first = m_backlinkTarget[f_node];
                m_bridges[f_node][0].second = m_backlinkSource[f_node];
                m_bridgeStatus[f_node] = 1;
            }
            m_intervalStores[f_node]->pop();
            m_intervalStores[f_node]->push(f_node);
        }
        m_bridges[m_intervalStores[f_node]->top()][m_bridgeStatus[m_intervalStores[f_node]->top()]].first = f_node;
        m_bridges[m_intervalStores[f_node]->top()][m_bridgeStatus[m_intervalStores[f_node]->top()]].second = f_parent;
        m_bridgeStatus[m_intervalStores[f_node]->top()]++;
    }
    return f_time;
}    


void DcjdDist::clearBridges()
{
    delete[] m_timestamp;
    delete[] m_numArrivingBacklinks;
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        delete[] m_arrivingBacklinks[i];
    delete[] m_arrivingBacklinks;
    delete[] m_numOpenBacklinks;
    delete[] m_backlinkSource;
    delete[] m_backlinkTarget;
    delete[] m_longestBacklink;
    delete[] m_intervalStacks;
    delete[] m_intervalStores;
    delete[] m_openInterval;
    delete[] m_timestamp2Node;
    delete[] m_bridgeStatus;
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        delete[] m_bridges[i];
    delete[] m_bridges;
    delete[] m_bridgeCompID;
    delete[] m_maxChild;
} 


int DcjdDist::listGoodMoves()
{
    // sorry this implementation will be rather slow...
    
    int numMoves = 0;   // this will be returned
    int v0, v1;         // two vertices 
    int v2, v3;         // another two vertices (needed for DCJ)
    int p0, p1, p2, p3; // positions of v0 .. v3 in the permutation 
    int el0, el1;       // two elements in the permutation
    int el2, el3;       // another two elements in the permutation (needed for DCJ)
    int pos = 0;        // position of the first element after the duplication
    bool dupl;          // true if a 1-bridge is a true duplication
    bool found;
    int left0, left1;   // left positions of the cuts in the perm
    int connectionType; // how to connect DCJ
    Operation* op;       // the current operation
    
    // init (avoid compiler warnings
    p0 = -1;
    p1 = -1;
    p2 = -1;
    p3 = -1;
    left0 = -1;
    left1 = -1;
    // clean up old vector
    for (unsigned int i = 0; i < m_goodMoves.size(); i++)
        delete m_goodMoves[i];
    m_goodMoves.clear();
    m_scores.clear();
    // check for duplications (1-bridges)
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        if (m_bridgeStatus[i] == 1)   
        {
            // check 1: is the edge a desire edge?
            if (m_bridges[i][0].first < m_bridges[i][0].second)
            {
                v0 = m_bridges[i][0].first;
                v1 = m_bridges[i][0].second;
            }
            else
            {
                v1 = m_bridges[i][0].first;
                v0 = m_bridges[i][0].second;
            }
            if ((v1 & 1) && (v1 == v0+1))
                continue;
            // check 2: is the duplication valid?
            if (v0 & 1)
                el0 = -(v0 / 2) - 1;
            else 
                el0 = v0 / 2;
            if (v1 & 1)
                el1 = v1 / 2 + 1; 
            else
                el1 = -(v1 / 2);
            for (int j = 0; j < m_permSize - 1; j++)
            {
                if (((m_perm[j] == el0) && (m_perm[j+1] == el1))
                  || ((m_perm[j] == -el1) && (m_perm[j+1] == -el0)))
                {  
                    pos = j+1;
                    break;
                }
            }
            dupl = false;
            for (int j = 0; j < pos; j++)
            {
                if (pos - j > m_permSize - pos)    
                    continue;
                found = true;    
                for (int k = 0; k < pos - j; k++)
                {
                    if (m_perm[j+k] != m_perm[pos+k])   // no duplication
                    {
                        found = false;
                        break;
                    }
                }
                if (found)            // true duplication
                {
                    op = new Operation();
                    op->m_type = Operation::OPERATION_DUPLICATION;
                    op->m_edges[0] = j;
                    op->m_edges[1] = pos;
                    m_goodMoves.push_back(op);
                    dupl = true;
                    numMoves++;
                }
            }
        }
    }
    // check for DCJ (2-bridges)
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        if (m_bridgeStatus[i] >= 2)   
        {
            for (int j = 0; j < m_bridgeStatus[i] - 1; j++)   // for all edges in the cycle (first of DCJ)
            {
                 // is this a desire edge or adjacency? then continue
                if (m_bridges[i][j].first < m_bridges[i][j].second)
                {
                    v0 = m_bridges[i][j].first;
                    v1 = m_bridges[i][j].second;
                }
                else
                {
                    v1 = m_bridges[i][j].first;
                    v0 = m_bridges[i][j].second;
                }
                if ((v1 & 1) && (v1 == v0+1))
                    continue;
                for (int k = j+1; k < m_bridgeStatus[i]; k++)   // for all edges in the cycle (second of DCJ)
                {
                     // is this a desire edge or adjacency? then continue
                    if (m_bridges[i][k].first < m_bridges[i][k].second)
                    {
                        v2 = m_bridges[i][k].first;
                        v3 = m_bridges[i][k].second;
                    }
                    else
                    {
                        v3 = m_bridges[i][k].first;
                        v2 = m_bridges[i][k].second;
                    }
                    if ((v3 & 1) && (v3 == v2+1))
                        continue;
                    // now we have a real DCJ; classify it!
                    // set positions v0, v1, v2, v3 to be cyclic in the graph
                    v0 = m_bridges[i][j].first;
                    v1 = m_bridges[i][j].second;
                    v2 = m_bridges[i][k].first;
                    v3 = m_bridges[i][k].second;
                    if (v0 & 1)
                        el0 = -(v0 / 2) - 1;
                    else 
                        el0 = v0 / 2;
                    if (v1 & 1)
                        el1 = v1 / 2 + 1; 
                    else
                        el1 = -(v1 / 2);
                    if (v2 & 1)
                        el2 = -(v2 / 2) - 1;
                    else 
                        el2 = v2 / 2;
                    if (v3 & 1)
                        el3 = v3 / 2 + 1; 
                    else
                        el3 = -(v3 / 2);
                    p0 = -1;    
                    p1 = -1;    
                    p2 = -1;    
                    p3 = -1;    
                    // calculate positions in pi
                    for (int l = 0; l < m_permSize - 1; l++)
                    {
                        if ((p0 == -1) && (m_perm[l] == el0) && (m_perm[l+1] == el1))
                        {
                            p0 = 2 * l;
                            p1 = 2 * l + 1;
                            left0 = l;
                        }
                        else if ((p0 == -1) && (m_perm[l] == -el1) && (m_perm[l+1] == -el0))
                        {
                            p0 = 2 * l + 1;
                            p1 = 2 * l;
                            left0 = l;
                        }
                        else if ((p2 == -1) && (m_perm[l] == el2) && (m_perm[l+1] == el3))
                        {
                            p2 = 2 * l;
                            p3 = 2 * l + 1;
                            left1 = l;
                        }
                        else if ((p2 == -1) && (m_perm[l] == -el3) && (m_perm[l+1] == -el2))
                        {
                            p2 = 2 * l + 1;
                            p3 = 2 * l;
                            left1 = l;
                        }
                    }
                    op = new Operation();
                    if (left0 < left1)
                    {
                        op->m_edges[0] = left0 + 1;
                        op->m_edges[1] = left1 + 1;
                        if (p0 == 2 * left0)      // connect to p3
                        {
                            if (p3 & 1)
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                        else                      // connect to p2
                        {
                            if (p2 & 1) 
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                    }
                    else
                    {
                        op->m_edges[0] = left1 + 1;
                        op->m_edges[1] = left0 + 1;
                        if (p2 == 2 * left1)      // connect to p1
                        {
                            if (p1 & 1)
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                        else                      // connect to p0
                        {
                            if (p0 & 1) 
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                    }
                    m_goodMoves.push_back(op);
                    numMoves++;
                }
            }
        }
    }
    // check for DCJ (2 1-bridges)
    for (int i = 0; i < m_numMarkers * 2 - 3; i++)
    {
        if (m_bridgeStatus[i] == 1)   
        {
            // check 1: is the edge a desire edge?
            if (m_bridges[i][0].first < m_bridges[i][0].second)
            {
                v0 = m_bridges[i][0].first;
                v1 = m_bridges[i][0].second;
            }
            else
            {
                v1 = m_bridges[i][0].first;
                v0 = m_bridges[i][0].second;
            }
            if ((v1 & 1) && (v1 == v0+1))             // desire edge
                continue;
            // we found a 1-bridge; check if we find another one in the same component    
            for (int j = i+1; j < m_numMarkers * 2 - 2; j++)
            {
                if (m_bridgeStatus[j] == 1) 
                {
                    // check 2: is this in the same component?
                    if (m_bridgeCompID[v0] != m_bridgeCompID[m_bridges[j][0].first])
                        continue;
                    // check 3: is this a desire edge?  
                    if (m_bridges[j][0].first < m_bridges[j][0].second)
                    {
                        v2 = m_bridges[j][0].first;
                        v3 = m_bridges[j][0].second;
                    }
                    else
                    {
                        v3 = m_bridges[j][0].first;
                        v2 = m_bridges[j][0].second;
                    }
                    if ((v3 & 1) && (v3 == v2+1))             // desire edge
                        continue;
                    // both 1-bridges are in the same component
                    // get out the connection
                    v0 = m_bridges[i][0].first;
                    v1 = m_bridges[i][0].second;
                    v2 = m_bridges[j][0].first;
                    v3 = m_bridges[j][0].second;
                    if (v0 & 1)
                        el0 = -(v0 / 2) - 1;
                    else 
                        el0 = v0 / 2;
                    if (v1 & 1)
                        el1 = v1 / 2 + 1; 
                    else
                        el1 = -(v1 / 2);
                    if (v2 & 1)
                        el2 = -(v2 / 2) - 1;
                    else 
                        el2 = v2 / 2;
                    if (v3 & 1)
                        el3 = v3 / 2 + 1; 
                    else
                        el3 = -(v3 / 2);
                    // calculate positions in pi
                    for (int l = 0; l < m_permSize - 1; l++)
                    {
                        if ((m_perm[l] == el0) && (m_perm[l+1] == el1))
                        {
                            p0 = 2 * l;
                            p1 = 2 * l + 1;
                            left0 = l;
                        }
                        else if ((m_perm[l] == -el1) && (m_perm[l+1] == -el0))
                        {
                            p0 = 2 * l + 1;
                            p1 = 2 * l;
                            left0 = l;
                        }
                        else if ((m_perm[l] == el2) && (m_perm[l+1] == el3))
                        {
                            p2 = 2 * l;
                            p3 = 2 * l + 1;
                            left1 = l;
                        }
                        else if ((m_perm[l] == -el3) && (m_perm[l+1] == -el2))
                        {
                            p2 = 2 * l + 1;
                            p3 = 2 * l;
                            left1 = l;
                        }
                    }
                    if ((m_timestamp[v0] <= m_timestamp[v3]) && (m_maxChild[v1] >= m_timestamp[v2]))
                        // path v1 - v0 - v3 - v2 -> connect v1 - v2, v0 - v3
                        connectionType = 0;
                    else if ((m_timestamp[v2] <= m_timestamp[v1]) && (m_maxChild[v3] >= m_timestamp[v0]))
                        // path v3 - v2 - v1 - v0 -> connect v0 - v3, v1 - v2
                        connectionType = 0;
                    else
                        // v1 - v0, v3 - v2 in different subtrees -> connect v1 - v3, v0 - v2
                        connectionType = 1;
                    op = new Operation();
                    if (left0 < left1)
                    {
                        op->m_edges[0] = left0 + 1;
                        op->m_edges[1] = left1 + 1;
                        if (((p0 == 2 * left0) && (connectionType == 0))        // connect to v3
                          || ((p1 == 2 * left0) && (connectionType == 1)))
                        {
                            if (p3 & 1)
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                        else                                                    // connect to v2
                        {
                            if (p2 & 1)           
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                    }
                    else
                    {
                        op->m_edges[0] = left1 + 1;
                        op->m_edges[1] = left0 + 1;
                        if (((p2 == 2 * left1) && (connectionType == 0))        // connect to v1
                          || ((p3 == 2 * left1) && (connectionType == 1)))
                        {
                            if (p1 & 1)
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                        else                                                     // connect to v0
                        {
                            if (p0 & 1)
                                op->m_type = Operation::OPERATION_DCJCIRC;
                            else
                                op->m_type = Operation::OPERATION_REVERSAL;
                        }
                    }
                    m_goodMoves.push_back(op);
                    numMoves++;        
                }
            }
        }
    }
    // checking for selfloops
    for (int i = 0; i < m_numMarkers * 2 - 3; i++)
    {
        if (m_numSelfloops[i] != 1)
            continue;
        if (i & 1)    
            el0 = -((i+1) / 2);
        else 
            el0 = i / 2;
        for (int j = 0; j < m_permSize; j++)
        {
            if ((m_perm[j] == el0) && (m_perm[j+1] == -el0))
            {
                p0 = j;
                break;
            }
        }     
        // check pair of selfloops
        for (int j = i+1; j < m_numMarkers * 2 - 2; j++)
        {
            if (m_numSelfloops[j] != 1)
                continue;
            if (j & 1)    
                el2 = -((j+1) / 2);
            else 
                el2 = j / 2;
            for (int k = 0; k < m_permSize; k++)
            {
                if ((m_perm[k] == el2) && (m_perm[k+1] == -el2))
                {
                    p2 = k;
                    break;
                }
            }     
            op = new Operation;
            op->m_type = Operation::OPERATION_REVERSAL;
            if (p0 < p2)
            {
                op->m_edges[0] = p0 + 1;
                op->m_edges[1] = p2 + 1;
            }
            else
            {
                op->m_edges[0] = p2 + 1;
                op->m_edges[1] = p0 + 1;
            }
            m_goodMoves.push_back(op);
            numMoves++;
        }
        // check selfloop with another edge of the same component
        // do all this just as reversals (maybe change when we have lookahead)
        for (int j = 0; j < 2 * m_numMarkers -2 ; j++)
        {
            if (i == j)
                continue;
            if (m_bridgeCompID[i] != m_bridgeCompID[j])
                continue;
            if (j & 1)    
                el2 = -((j+1) / 2);
            else 
                el2 = j / 2;
            for (int k = 0; k < m_rddDegree[j]; k++)
            {
                if (m_rddEdges[j][k] == j)   // selfloop -> already done
                    continue;
                if (m_rddEdges[j][k] == i)   // leads to selfloop
                    continue;
                if (((j & 1) && (m_rddEdges[j][k] == j-1))   // adjacency
                  || (!(j & 1) && (m_rddEdges[j][k] == j+1)))
                    continue;
                if (m_rddEdges[j][k] & 1)
                    el3 = (m_rddEdges[j][k]+1) / 2;
                else
                    el3 = -(m_rddEdges[j][k] / 2);
                for (int l = 0; l < m_permSize; l++)
                {
                    if (((m_perm[l] == el2) && (m_perm[l+1] == el3)) 
                      || ((m_perm[l] == -el3) && (m_perm[l+1] == -el2)))
                    {
                        p2 = l;
                        break;
                    }
                }
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                if (p0 < p2)
                {
                    op->m_edges[0] = p0 + 1;
                    op->m_edges[1] = p2 + 1;
                }
                else
                {
                    op->m_edges[0] = p2 + 1;
                    op->m_edges[1] = p0 + 1;
                }
                m_goodMoves.push_back(op);
                numMoves++;
            }
        }
        // check if selfloop is a tandem duplication
        for (int j = (2 * p0 + 2 >= m_permSize? 2 * p0 + 2 - m_permSize : 0); j <= p0; j++)
        {
            p1 = j;
            p2 = p0 + 1;
            while ((p1 <= p0) && (m_perm[p1] == m_perm[p2]))
            {
                p1++;
                p2++;
            }
            if (p1 > p0)         // duplication found
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = j;
                op->m_edges[1] = p0 + 1;
                m_goodMoves.push_back(op);
                numMoves++;
            }
        }                
    }
    // checking for deletions
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        if (m_bridgeStatus[i] == 1)   
        {
            // check 1: is the edge a desire edge?
            if (m_bridges[i][0].first < m_bridges[i][0].second)
            {
                v0 = m_bridges[i][0].first;
                v1 = m_bridges[i][0].second;
            }
            else
            {
                v1 = m_bridges[i][0].first;
                v0 = m_bridges[i][0].second;
            }
            if ((v1 & 1) && (v1 == v0+1))             // desire edge
                continue;
            // check 2: find the edge to split
            if (v0 & 1)
                el0 = -(v0 / 2) - 1;
            else 
                el0 = v0 / 2;
            if (v1 & 1)
                el1 = v1 / 2 + 1; 
            else
                el1 = -(v1 / 2);
            for (int j = 0; j < m_permSize - 1; j++)
            {
                if (((m_perm[j] == el0) && (m_perm[j+1] == el1))
                  || ((m_perm[j] == -el1) && (m_perm[j+1] == -el0)))
                {  
                    pos = j+1;
                    break;
                }
            }
            op = findDeletion3(pos);
            if (op != NULL)
            {
                m_goodMoves.push_back(op);
                numMoves++;
            }
        }
    }
        
    return numMoves;
} 
       

int DcjdDist::getAllFollowUps(Operation& f_op)
{
    int numFollowUps = 0;
    Operation* op;
    
    for (int i = 1; i < f_op.m_edges[0]; i++)
    {
        for (int j = f_op.m_edges[0]; j < f_op.m_edges[1]; j++)
        {
            op = new Operation;
            op->m_type = Operation::OPERATION_BLOCKINTERCHANGE;
            op->m_edges[0] = i;
            op->m_edges[1] = f_op.m_edges[0];
            op->m_edges[2] = j;
            op->m_edges[3] = f_op.m_edges[1];
            m_goodMoves.push_back(op);
            numFollowUps++;
        }
    }
    for (int i = f_op.m_edges[0] + 1; i <= f_op.m_edges[1]; i++)
    {
        for (int j = f_op.m_edges[1] + 1; j < m_permSize; j++)
        {
            op = new Operation;
            op->m_type = Operation::OPERATION_BLOCKINTERCHANGE;
            op->m_edges[0] = f_op.m_edges[0];
            op->m_edges[1] = i;
            op->m_edges[2] = f_op.m_edges[1];
            op->m_edges[3] = j;
            m_goodMoves.push_back(op);
            numFollowUps++;
        }
    }
    return numFollowUps;    
}


void DcjdDist::findReverseDuplications()
{
    int length;          // length of the duplication
    Operation* op;       // the current operation
    vector<Operation*> seq;  // sequence of operations
    
    for (int i = 0; i < m_permSize - 1; i++)
    {
        length = 0;
        while ((length <= i) && (i + length + 1 < m_permSize))
        { 
            if (m_perm[i-length] == -m_perm[i+1+length])
                length++;
            else 
                break;
        }        
        if (length > 0)
        {
            seq.clear();
            op = new Operation();
            op->m_type = Operation::OPERATION_REVERSAL;
            op->m_edges[0] = i - length + 1;
            op->m_edges[1] = i + 1;
            seq.push_back(op);
            op = new Operation();
            op->m_type = Operation::OPERATION_DUPLICATION;
            op->m_edges[0] = i - length + 1;
            op->m_edges[1] = i + 1;
            seq.push_back(op);
            m_sequences.push_back(seq);
            seq.clear();
            op = new Operation();
            op->m_type = Operation::OPERATION_REVERSAL;
            op->m_edges[0] = i + 1;
            op->m_edges[1] = i + 1 + length;
            seq.push_back(op);
            op = new Operation();
            op->m_type = Operation::OPERATION_DUPLICATION;
            op->m_edges[0] = i - length + 1;
            op->m_edges[1] = i + 1;
            seq.push_back(op);
            m_sequences.push_back(seq);
        }
    }
}

 
Operation* DcjdDist::findDeletion3(int f_pos)
{
    Operation* result = NULL;
    int e0, e1;            // elements around the cut point
    int missing = 0;       // number of missing elements between e0 and e1
    bool found;            // true if we found a specific element
    
    e0 = m_perm[f_pos-1];
    e1 = m_perm[f_pos];
    // check if we can fill with adjacencies, count missing elements
    if (((e0 >= 0) && (e1 >= 0) && (e0 < e1))
     || ((e0 < 0) && (e1 < 0) && (e0 < e1)))
    {
        for (int i = e0 + 1; i < e1; i++)
        {    
            found = false;
            for (int j = 0; j < m_permSize; j++)
            {
                if ((m_perm[j] == i) || (m_perm[j] == -i))
                {
                    found = true;
                    break;
                }
            }
            if (!found)
            {
                missing++;
                break;
            }
        }
    }
    
/*    if ((result != NULL) && ((int)(result->m_sequence.size()) > m_permSize / 10) 
          && (m_permSize > m_numMarkers))
    {  
        delete result;
        result = NULL;
    }*/
    
    // check the scores
    if (!(((e0 >= 0) && (e1 >= 0) && (e0 < e1))
      || ((e0 < 0) && (e1 < 0) && (e0 < e1))))
        return result;      // either type 2 or NULL
    else if ((result == NULL) || (3 * missing - (e1 - e0 - 1) > (int)(result->m_sequence.size())))
    {
        if (result != NULL)
            delete result;
        result = new Operation;
        result->m_type = Operation::OPERATION_DELETION;
        result->m_edges[0] = f_pos;
        for (int i = e0 + 1; i < e1; i++)
            result->m_sequence.push_back(i);
    }
    if (result != NULL)
        return result; 
    return NULL;    
}


int DcjdDist::findAdjacencySequences()
{
    unsigned int oldSize = m_sequences.size();
    int t0, t1;              // targets of reality edges
    int e0, e1, e2, e3;      // elements in the permutation
    int p0, p1, p2, p3;      // positions in the permutation
    int or0, or1;            // orientation of the adjacencies
    Operation* op;           // an operation of a sequence
    vector<Operation*> seq;  // sequence of operations
    
    // init
    or0 = 0;
    or1 = 0;
    p0 = -1;
    p1 = -1;
    p2 = -1;
    p3 = -1;
    // find possible configurations
    for (int i = 0; i < m_numMarkers - 1; i++)
    {
        for (int j = 0; j < m_rddDegree[2*i]; j++)
        {
            t0 = m_rddEdges[2*i][j];
            if (t0 == 2*i+1)
                continue;
            for (int k = 0; k < m_rddDegree[2*i+1]; k++)
            {
                t1 = m_rddEdges[2*i+1][k];
                if ((t1 == 2*i) || (t1 == t0))
                    continue;
                // if we are here, reality edges (2*i, t0) 
                // and (2*i+1, t1) can build an adjacency
                // find those elements in the permutation
                e0 = i;
                if (t0 & 1)
                    e1 = (t0 + 1) / 2;
                else
                    e1 = -t0 / 2;
                e2 = i + 1;
                if (t1 & 1)
                    e3 = -(t1 + 1) / 2;
                else
                    e3 = t1 / 2;
                // we have now the adjacencies e0 e1 and e3 e2    
                for (int l = 0; l < m_permSize - 1; l++)
                {
                    if ((m_perm[l] == e0) && (m_perm[l+1] == e1))
                    {
                        p0 = l;
                        p1 = l + 1;
                        or0 = 1;
                        break;
                    }
                    if ((m_perm[l+1] == -e0) && (m_perm[l] == -e1))
                    {
                        p0 = l + 1;
                        p1 = l;
                        or0 = -1;
                        break;
                    }
                }
                for (int l = 0; l < m_permSize - 1; l++)
                {
                    if ((m_perm[l] == e3) && (m_perm[l+1] == e2))
                    {
                        p3 = l;
                        p2 = l + 1;
                        or1 = 1;
                        break;
                    }
                    if ((m_perm[l+1] == -e3) && (m_perm[l] == -e2))
                    {
                        p3 = l + 1;
                        p2 = l;
                        or1 = -1;
                        break;
                    }
                }
                if (or0 == or1)            // would cause CI
                {
                    if (or0 == 1)
                    {
                        for (int l = p0; l < m_permSize - 1; l++)
                        {
                            if ((m_perm[l+1] == -e0) && (m_perm[l] == -e1))
                            {
                                p0 = l+1;
                                p1 = l;
                                or0 = -1;
                                break;
                            }
                        }
                    }
                    else
                    {
                        for (int l = p1; l < m_permSize - 1; l++)
                        {
                            if ((m_perm[l] == e0) && (m_perm[l+1] == e1))
                            {
                                p0 = l;
                                p1 = l+1;
                                or0 = 1;
                                break;
                            }
                        }
                    }
                }
                if (or0 == or1)            // would cause CI
                {
                    if (or1 == 1)
                    {
                        for (int l = p2; l < m_permSize - 1; l++)
                        {
                            if ((m_perm[l+1] == -e3) && (m_perm[l] == -e2))
                            {
                                p3 = l+1;
                                p2 = l;
                                or1 = -1;
                                break;
                            }
                        }
                    }
                    else
                    {
                        for (int l = p3; l < m_permSize - 1; l++)
                        {
                            if ((m_perm[l] == e3) && (m_perm[l+1] == e2))
                            {
                                p3 = l;
                                p2 = l+1;
                                or1 = 1;
                                break;
                            }
                        }
                    }
                }
                if (or0 == or1)    
                    continue;
                seq.clear();
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                if (or0 == 1)
                {
                    if (p0 < p2)
                    {
                        op->m_edges[0] = p1;
                        op->m_edges[1] = p3;
                    }
                    else
                    {
                        op->m_edges[0] = p3;
                        op->m_edges[1] = p1;
                    }
                }
                else
                {
                    if (p1 < p2)
                    {
                        op->m_edges[0] = p0;
                        op->m_edges[1] = p2;
                    }
                    else
                    {
                        op->m_edges[0] = p2;
                        op->m_edges[1] = p0;
                    }
                }
                seq.push_back(op);
                m_sequences.push_back(seq);   
            }
        }
    }
    return m_sequences.size() - oldSize;
}


int DcjdDist::findType2Sequences()
{
    unsigned int oldSize = m_sequences.size();
    int source;              // source node of the reality edges going to node a
    int t0, t1;              // target of the desire edges
    int conftype;            // type of the configuration; 0: A; 1: B; 2: C; -1: handle later
    int e0, e1, e2, e3;      // elements in the permutation
    int p0, p1, p2, p3;      // positions in the permutation
    int value;               // an arbitrary value (element)
    vector<Operation*> seq;  // sequence of operations
    Operation* op;           // an operation of the sequence
    int or0, or1;            // orientation of two segments (+1: positive, -1: negative)
    int dupllength;          // length of the duplication
    
    // init (avoid compiler warnings)
    p0 = -1;
    p1 = -1;
    p2 = -1;
    p3 = -1;
    e0 = 0;
    e1 = 0;
    e2 = 0;
    e3 = 0;
    t0 = -1;
    t1 = -1;
    or0 = 0;
    or1 = 1;
    // find possible configurations
    for (int i = 0; i < m_numMarkers - 1; i++)
    {
        conftype = -1;
        if (m_rddDegree[2*i] == 1)
        {
            if (m_rddDegree[2*i+1] == 2)
            {
                source = 2*i+1;
                if (m_rddEdges[2*i+1][0] == 2*i)
                    t0 = m_rddEdges[2*i+1][1];
                else    
                    t0 = m_rddEdges[2*i+1][0];
                conftype = 0;    
            }
            else if (m_rddDegree[2*i+1] == 3)
            {
                source = 2*i+1;
                if (m_rddEdges[2*i+1][0] == 2*i)
                {
                    t0 = m_rddEdges[2*i+1][1];
                    t1 = m_rddEdges[2*i+1][2];
                }
                else if (m_rddEdges[2*i+1][1] == 2*i)
                {
                    t0 = m_rddEdges[2*i+1][0];
                    t1 = m_rddEdges[2*i+1][2];
                }
                else
                {
                    t0 = m_rddEdges[2*i+1][0];
                    t1 = m_rddEdges[2*i+1][1];
                }
                if (t0 != t1)
                    conftype = 2;
            }
        }
        else if (m_rddDegree[2*i] == 2)
        {
            if (m_rddDegree[2*i+1] == 1)
            { 
                source = 2*i;
                if (m_rddEdges[2*i][0] == 2*i+1)
                    t0 = m_rddEdges[2*i][1];
                else
                    t0 = m_rddEdges[2*i][0];
                conftype = 0;
            }
            else if (m_rddDegree[2*i+1] == 3)
            {
                if (m_rddEdges[2*i][0] == m_rddEdges[2*i][1])
                {
                    source = 2*i+1;
                    if (m_rddEdges[2*i+1][0] != 2*i)
                        t0 = m_rddEdges[2*i+1][0];
                    else if (m_rddEdges[2*i+1][1] != 2*i)
                        t0 = m_rddEdges[2*i+1][1];
                    else
                        t0 = m_rddEdges[2*i+1][2];
                    conftype = 1;
                }
            }
        }
        else if (m_rddDegree[2*i] == 3)
        {
            if (m_rddDegree[2*i+1] == 1)
            {
                source = 2*i;
                if (m_rddEdges[2*i][0] == 2*i+1)
                {
                    t0 = m_rddEdges[2*i][1];
                    t1 = m_rddEdges[2*i][2];
                }
                else if (m_rddEdges[2*i][1] == 2*i+1)
                {
                    t0 = m_rddEdges[2*i][0];
                    t1 = m_rddEdges[2*i][2];
                }
                else
                {
                    t0 = m_rddEdges[2*i][0];
                    t1 = m_rddEdges[2*i][1];
                }
                if (t0 != t1)
                    conftype = 2;
            }
            else if (m_rddDegree[2*i+1] == 2)
            {
                if (m_rddEdges[2*i+1][0] == m_rddEdges[2*i+1][1])
                {
                    source = 2*i;
                    if (m_rddEdges[2*i][0] != 2*i+1)
                        t0 = m_rddEdges[2*i][0];
                    else if (m_rddEdges[2*i][1] != 2*i+1)
                        t0 = m_rddEdges[2*i][1];
                    else
                        t0 = m_rddEdges[2*i][2];
                    conftype = 1;
                }
            }
        }
        if (conftype == -1)
            continue;
        if (source & 1)
            e0 = -((source+1) / 2);
        else
            e0 = source / 2;
        if (t0 & 1)
            e1 = (t0+1) / 2;
        else
            e1 = -(t0 / 2);
        for (int i = 0; i < m_permSize - 1; i++)
        {
            if ((m_perm[i] == e0) && (m_perm[i+1] == e1))
            {
                p0 = i;
                p1 = i+1;
                break;
            }
            if ((m_perm[i+1] == -e0) && (m_perm[i] == -e1))
            {
                p0 = i+1;
                p1 = i;
                break;
            }
        }
        if (conftype == 0)
        {
            seq.clear();
            op = new Operation;
            op->m_type = Operation::OPERATION_DELETION;
            op->m_edges[0] = (p0 > p1? p0 : p1);
            e0 = abs(e0);  
            if (source & 1)
            {
                // make value the largest element < e0 that is not missing
                value = source - 1;
                while (m_rddDegree[value] == 1)
                    value -= 2;
                value /= 2;
                op->m_sequence.reserve(e0 - value - 1);
                if (m_perm[p0] >= 0)
                {
                    for (int j = value + 1; j < e0; j++)
                        op->m_sequence.push_back(j);
                }
                else
                {
                    for (int j = e0 - 1; j > value; j--)
                        op->m_sequence.push_back(-j);
                } 
            }
            else 
            {
                // make value the smallest element > e0 that is not missing
                value = source + 2;
                while (m_rddDegree[value] == 1)
                    value += 2;
                value /= 2;
                op->m_sequence.reserve(value - e0 - 1);
                if (m_perm[p0] >= 0)
                {
                    for (int j = e0 + 1; j < value; j++)
                        op->m_sequence.push_back(j);
                }
                else
                {
                    for (int j = value - 1; j > e0; j--)
                        op->m_sequence.push_back(-j);
                }
            }
            seq.push_back(op);
            m_sequences.push_back(seq);
        }
        else if (conftype == 1)
        {
            if (p0 > p1)
                or0 = 1;
            else
                or0 = -1;            
            // find the edge (x-1, x)
            e2 = e0;
            e3 = e2 + 1;
            for (int j = 0; j < m_permSize - 1; j++)
            {
                if ((m_perm[j] == e2) && (m_perm[j+1] == e3))
                {
                    p2 = j;
                    or1 = -1;
                    break;
                } 
                if ((m_perm[j+1] == -e2) && (m_perm[j] == -e3))
                {
                    p2 = j+1;
                    or1 = 1;
                    break;
                }
            }
            // maximize the length of the segment
            p1 = p0;
            p3 = p2;
            while (m_perm[p1] * or0 == m_perm[p3] * or1)
            {
                p1 += or0;
                p3 += or1;
                if (p3 == p1 - or0) 
                    break;
            }                  
            // p1, p3 are now the elements beyond the segment
            dupllength = (p1 - p0) * or0;
            seq.clear();            
            // check now which subcase has to be applied
            if (or0 == or1)      // Transposition + Tandem duplication, maybe additional reversal
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_DELETION;
                if (p0 < p2)
                {
                    if (p0 < p1)
                    {
                        op->m_edges[0] = p0;
                        op->m_sequence.reserve(p2 - p1);
                        for (int i = p1; i < p2; i++)
                            op->m_sequence.push_back(m_perm[i]);
                        seq.push_back(op);
                        dupllength = p1 - p0 + op->m_sequence.size();
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = p0;
                        op->m_edges[1] = p0 + dupllength;
                        seq.push_back(op);
                    }
                    else
                    {
                        op->m_edges[0] = p1 + 1;
                        op->m_sequence.reserve(p3 - p0);
                        for (int i = p0 + 1; i < p3 + 1; i++)
                            op->m_sequence.push_back(m_perm[i]);
                        seq.push_back(op);
                        dupllength = p0 - p1 + op->m_sequence.size();
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = p1 + 1;
                        op->m_edges[1] = p1 + 1 + dupllength;
                        seq.push_back(op);
                    }
                }
                else
                {
                    if (p2 < p3)
                    {
                        op->m_edges[0] = p1;
                        op->m_sequence.reserve(p0 - p3);
                        for (int i = p3; i < p0; i++)
                            op->m_sequence.push_back(m_perm[i]);
                        seq.push_back(op);
                        dupllength = p3 - p2 + op->m_sequence.size();
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = p2;
                        op->m_edges[1] = p2 + dupllength;
                        seq.push_back(op);
                    }
                    else
                    {
                        op->m_edges[0] = p0 + 1;
                        op->m_sequence.reserve(p1 - p2);
                        for (int i = p2 + 1; i < p1 + 1; i++)
                            op->m_sequence.push_back(m_perm[i]);
                        seq.push_back(op);
                        dupllength = p2 - p3 + op->m_sequence.size();
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = p3 + 1;
                        op->m_edges[1] = p3 + 1 + dupllength;
                        seq.push_back(op);
                    }
                }
                // if this creates a selfloop, remove it
                if (m_perm[p0-or0] == -m_perm[p1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    if (p0 < p2)
                    {
                        if (p0 < p1)
                        {
                            op->m_edges[0] = p0;
                            op->m_edges[1] = p2 - dupllength;
                        }
                        else
                        {
                            op->m_edges[0] = p1 + 1;
                            op->m_edges[1] = p3 + 1 - dupllength;
                        }
                    }
                    else
                    {
                        if (p2 < p3)
                        {
                            op->m_edges[0] = p3;
                            op->m_edges[1] = p0;
                        }
                        else
                        {
                            op->m_edges[0] = p3 + 1;
                            op->m_edges[1] = p1 + 1;
                        }
                    }
                    seq.push_back(op);
                }
            }
            else if (p3 == p1 - or0)   // reverse tandem duplication
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                if (p0 < p2)
                {
                    op->m_edges[0] = p0;
                    op->m_edges[1] = p1;
                }
                else
                {
                    op->m_edges[0] = p1 + 1;
                    op->m_edges[1] = p0 + 1;
                }
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                if (p0 < p2)
                {
                    op->m_edges[0] = p0;
                    op->m_edges[1] = p1;
                }
                else
                {
                    op->m_edges[0] = p3 + 1;
                    op->m_edges[1] = p2 + 1;
                }
                seq.push_back(op);
            }
            else                 // Reversal + Tandem Duplication, maybe additional reversal
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                if (p0 < p2)
                {
                    if (p0 < p1)
                    {
                        op->m_edges[0] = p0;
                        op->m_edges[1] = p3 + 1;
                    }
                    else
                    {
                        op->m_edges[0] = p0 + 1;
                        op->m_edges[1] = p3;
                    }
                }
                else
                {
                    if (p2 < p3)
                    {
                        op->m_edges[0] = p3;
                        op->m_edges[1] = p0 + 1;
                    }
                    else
                    {
                        op->m_edges[0] = p3 + 1;
                        op->m_edges[1] = p0;
                    }
                } 
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                if (p0 < p2)
                {
                    if (p0 < p1)
                        op->m_edges[0] = p3 + 1 - dupllength;
                    else 
                        op->m_edges[0] = p1 + 1;
                }
                else
                {
                    if (p2 < p3)
                        op->m_edges[0] = p2;
                    else
                        op->m_edges[0] = p0 - dupllength;
                }
                op->m_edges[1] = op->m_edges[0] + dupllength;
                seq.push_back(op);
                // check if we created a selfloop
                if (m_perm[p0-or0] == -m_perm[p3])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    if (p0 < p2)
                    {
                        if (p0 < p1)
                        {
                            op->m_edges[0] = p0;
                            op->m_edges[1] = p3 + 1 - dupllength;
                        }
                        else
                        {
                            op->m_edges[0] = p1 + 1;
                            op->m_edges[1] = p3 - dupllength;
                        }
                    }
                    else
                    {
                        if (p2 < p3)
                        {
                            op->m_edges[0] = p3;
                            op->m_edges[1] = p0 + 1 - dupllength;
                        }
                        else
                        {
                            op->m_edges[0] = p3 + 1;
                            op->m_edges[1] = p1 - dupllength;
                        }
                    }
                    seq.push_back(op);
                }
            }
            m_sequences.push_back(seq);                
        }
        else if (conftype == 2)
        {
// Bah damn CIs! Use the code for conftype 0, if this merges components there is 
// a subsequent deletion that splits those components again
            seq.clear();
            op = new Operation;
            op->m_type = Operation::OPERATION_DELETION;
            op->m_edges[0] = (p0 > p1? p0 : p1);  
            if (source & 1)
            {
                // make value the largest element < e0 that is not missing
                value = source - 1;
                while (m_rddDegree[value] == 1)
                    value -= 2;
                value /= 2;
                if (e0 > 0)
                {
                    op->m_sequence.reserve(e0 - value - 1);
                    for (int j = value + 1; j < e0; j++)
                        op->m_sequence.push_back(j);
                }
                else
                {
                    op->m_sequence.reserve(-e0 - value - 1);
                    for (int j = -e0 - 1; j > value; j--)
                        op->m_sequence.push_back(-j);
                } 
            }
            else 
            {
                // make value the smallest element > e0 that is not missing
                value = source + 2;
                while (m_rddDegree[value] == 1)
                    value += 2;
                value /= 2;
                if (e0 > 0)
                {
                    op->m_sequence.reserve(value - e0 - 1);
                    for (int j = e0 + 1; j < value; j++)
                        op->m_sequence.push_back(j);
                }
                else
                {
                    op->m_sequence.reserve(value + e0 - 1);
                    for (int j = value - 1; j > -e0; j--)
                        op->m_sequence.push_back(-j);
                }
            }
            seq.push_back(op);
            m_sequences.push_back(seq);
        } // end if conftype == 2
    }
    return m_sequences.size() - oldSize;
}    


int DcjdDist::findType3Sequences()
{
    unsigned int oldSize = m_sequences.size();
    vector<int> startpos;           // startpositions of the segments
    vector<int> ori;                // orientation of the segments (-1 or +1)
    int seglength;                  // length of the segments
    vector<bool> validSegments;     // true at each position where the segment is not yet discarded
    int score0, score1, score2;     // scores of different sequences
    vector<Operation*> seq;         // a sequence of operations
    Operation* op;                  // an operation of the sequence
    int v0, v1, v2, v3;             // vertices
    vector<int> spbak;              // backup of startpos
    vector<int> oribak;             // backup of ori
    
    for (int i = 1; i < m_numMarkers - 1; i++)
    {
        if (m_rddDegree[2*i] >= 4)       // at least 3 occurancies
        {
            startpos.clear();
            ori.clear();
            validSegments.clear();
            for (int j = 1; j < m_permSize; j++)     // search segment startpositions
            {
                if (m_perm[j] == i)
                {
                    startpos.push_back(j);
                    ori.push_back(1);
                    validSegments.push_back(true);
                }
                else if (m_perm[j] == -i)
                {
                    startpos.push_back(j);
                    ori.push_back(-1);
                    validSegments.push_back(true);
                }
            }
            
            spbak.clear();
            oribak.clear();
            for (unsigned int j = 0; j < startpos.size(); j++)
                spbak.push_back(startpos[j]);
            for (unsigned int j = 0; j < ori.size(); j++)
                oribak.push_back(ori[j]);
            for (unsigned int s0 = 0; s0 < startpos.size() - 2; s0++)
            {
                startpos[0] = spbak[s0];
                ori[0] = oribak[s0];
                for (unsigned int s1 = s0 + 1; s1 < startpos.size() - 1; s1++)
                {
                    startpos[1] = spbak[s1];
                    ori[1] = oribak[s1];
                    for (unsigned int s2 = s1 + 1; s2 < startpos.size(); s2++)
                    {
                        startpos[2] = spbak[s2];
                        ori[2] = oribak[s2];
                        // if we can extend to the left, discard
                        if ((m_perm[startpos[0] - ori[0]] * ori[0] == m_perm[startpos[1] - ori[1]] * ori[1])
                          && (m_perm[startpos[0] - ori[0]] * ori[0] == m_perm[startpos[2] - ori[2]] * ori[2]))
                            continue;
                        // extend to the right
                        seglength = 1;
                        while ((m_perm[startpos[0] + ori[0] * seglength] * ori[0] == m_perm[startpos[1] + ori[1] * seglength] * ori[1])
                          && (m_perm[startpos[0] + ori[0] * seglength] * ori[0] == m_perm[startpos[2] + ori[2] * seglength] * ori[2])
                          && (startpos[0] + ori[0] * seglength < startpos[1]) 
                          && (startpos[0] + ori[0] * seglength < startpos[1] + ori[1] * seglength)
                          && (startpos[0] < startpos[1] + ori[1] * seglength)
                          && (startpos[1] + ori[1] * seglength < startpos[2]) 
                          && (startpos[1] + ori[1] * seglength < startpos[2] + ori[2] * seglength)
                          && (startpos[1] < startpos[2] + ori[2] * seglength))
                            seglength++;
                        // tandem duplications are handled elsewhere    
                        if (((ori[0] == ori[1]) && (startpos[0] + seglength == startpos[1]))
                          || ((ori[1] == ori[2]) && (startpos[1] + seglength == startpos[2])))
                            continue;
            seq.clear();    
            // choose best strategy to delete two segments, and create appropriate sequence
            score0 = 0;
            score1 = 0;
            score2 = 0;
            if ((ori[0] == ori[1]) && (ori[0] == ori[2]))
            {
                if (m_perm[startpos[0] - ori[0]] + ori[0] == m_perm[startpos[0]])  // adjacency before seg0
                    score0++;
                if (m_perm[startpos[1] - ori[1]] + ori[1] == m_perm[startpos[1]])  // adjacency before seg1
                    score1++;
                if (m_perm[startpos[2] - ori[2]] + ori[2] == m_perm[startpos[2]])  // adjacency before seg2
                    score2++;
                if (m_perm[startpos[0] + ori[0] * seglength - ori[0]] + ori[0] == m_perm[startpos[0] + ori[0] * seglength])  // adj after seg0
                    score0++;    
                if (m_perm[startpos[1] + ori[1] * seglength - ori[1]] + ori[1] == m_perm[startpos[1] + ori[1] * seglength])  // adj after seg1
                    score1++;    
                if (m_perm[startpos[2] + ori[2] * seglength - ori[2]] + ori[2] == m_perm[startpos[2] + ori[2] * seglength])  // adj after seg0
                    score2++;    
                // check for possible selfloops
                if (m_perm[startpos[0] - ori[0]] == m_perm[startpos[1] - ori[1]])
                    score2 -= 5;
                if (m_perm[startpos[0] - ori[0]] == m_perm[startpos[2] - ori[2]])
                    score1 -= 5;
                if (m_perm[startpos[1] - ori[1]] == m_perm[startpos[2] - ori[2]])
                    score0 -= 5;
                if (m_perm[startpos[0] + ori[0] * seglength] == m_perm[startpos[1] + ori[1] * seglength])
                    score2 -= 5;    
                if (m_perm[startpos[0] + ori[0] * seglength] == m_perm[startpos[2] + ori[2] * seglength])
                    score1 -= 5;    
                if (m_perm[startpos[1] + ori[1] * seglength] == m_perm[startpos[2] + ori[2] * seglength])
                    score0 -= 5;
                if ((score0 >= score1) && (score0 >= score2))
                {
                    // create aAb ... ce ... df
                    if (ori[0] == 1)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[1] + seglength;
                        op->m_edges[1] = startpos[2] + seglength;
                        seq.push_back(op);
                        op =  new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = startpos[1] + 2 * seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = startpos[0] + 2 * seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + seglength;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] + seglength - 1] >= 0? 2 * m_perm[startpos[0] + seglength - 1]    
                          : -2 * m_perm[startpos[0] + seglength] - 1);     // last point of A
                        v1 = (m_perm[startpos[0] + seglength] >= 0? 2 * m_perm[startpos[0] + seglength] - 1
                          : -2 * m_perm[startpos[0] + seglength]);         // b
                        v2 = (m_perm[startpos[1] - 1] >= 0? 2 * m_perm[startpos[1] - 1]
                          : -2 * m_perm[startpos[1] - 1] - 1);             // c  
                        v3 = (m_perm[startpos[2] - 1] >= 0? 2 * m_perm[startpos[2] - 1]
                          : -2 * m_perm[startpos[2] - 1] - 1);             // e  
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v2]) || (v0 == v2) || (v1 == v3))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[0] + seglength;
                            op->m_edges[1] = startpos[1];
                            seq.push_back(op);
                        }
                    }
                    else  // orientations negative
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[1] - seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[2] - 2 * seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] - seglength + 1] >= 0? 2 * m_perm[startpos[0] - seglength + 1] - 1
                          : -2 * m_perm[startpos[0] - seglength + 1]);     // last point of A
                        v1 = (m_perm[startpos[0] - seglength] >= 0? 2 * m_perm[startpos[0] - seglength]
                          : -2 * m_perm[startpos[0] - seglength] - 1);     // b
                        v2 = (m_perm[startpos[1] + 1] >= 0? 2 * m_perm[startpos[1] + 1] - 1
                          : -2 * m_perm[startpos[1] + 1]);                 // c
                        v3 = (m_perm[startpos[2] + 1] >= 0? 2 * m_perm[startpos[2] + 1] - 1
                          : -2 * m_perm[startpos[2] + 1]);                 // e
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v2]) || (v0 == v3) || (v1 == v2))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[0] - seglength + 1;
                            op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                            seq.push_back(op);
                        }
                    }
                }        
                else if (score1 >= score2)
                {
                    // create cAd ... ae ... bf
                    if (ori[0] == 1)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = startpos[2] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1];
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] + seglength - 1] >= 0? 2 * m_perm[startpos[0] + seglength - 1]
                          : -2 * m_perm[startpos[0] + seglength - 1] - 1);     // last point of A
                        v1 = (m_perm[startpos[0] - 1] >= 0? 2 * m_perm[startpos[0] - 1]
                          : -2 * m_perm[startpos[0] - 1] - 1);                 // a
                        v2 = (m_perm[startpos[1] + seglength] >= 0? 2 * m_perm[startpos[1] + seglength] - 1
                          : -2 * m_perm[startpos[1] + seglength]);             // d
                        v3 = (m_perm[startpos[2] - 1] >= 0? 2 * m_perm[startpos[2] - 1]
                          : -2 * m_perm[startpos[2] - 1] - 1);                 // e
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v1]) || (v0 == v3) || (v1 == v2))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[0];
                            op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] - seglength;
                            seq.push_back(op);
                        }
                    }
                    else     // orientations negative
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength + 1;
                        op->m_edges[1] = startpos[2] + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - 2 * seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - 2 * seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] - seglength + 1] >= 0? 2 * m_perm[startpos[0] - seglength + 1] - 1
                          : -2 * m_perm[startpos[0] - seglength + 1]);     // last point of A
                        v1 = (m_perm[startpos[0] + 1] >= 0? 2 * m_perm[startpos[0] + 1] - 1
                          : -2 * m_perm[startpos[0] + 1]);                 // a
                        v2 = (m_perm[startpos[1] - seglength] >= 0? 2 * m_perm[startpos[1] - seglength]
                          : -2 * m_perm[startpos[1] - seglength] - 1);     // d
                        v3 = (m_perm[startpos[2] + 1] >= 0? 2 * m_perm[startpos[2] + 1] - 1
                          : -2 * m_perm[startpos[2] + 1]);                 // e
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v1]) || (v0 == v1) || (v2 == v3))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength + 1;
                            op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                            seq.push_back(op);
                        }
                    }
                }
                else
                {
                    if (ori[0] == 1)
                    {
                        // create eAf ... ac ... bd
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = startpos[1] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + 2 * seglength;
                        op->m_edges[1] = startpos[2] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + seglength;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] + seglength - 1] >= 0? 2 * m_perm[startpos[0] + seglength - 1]
                          : -2 * m_perm[startpos[0] + seglength - 1] - 1);     // last point of A
                        v1 = (m_perm[startpos[0] - 1] >= 0? 2 * m_perm[startpos[0] - 1]
                          : -2 * m_perm[startpos[0] - 1] - 1);                 // a
                        v2 = (m_perm[startpos[1] - 1] >= 0? 2 * m_perm[startpos[1] - 1]
                          : -2 * m_perm[startpos[1] - 1] - 1);                 // c
                        v3 = (m_perm[startpos[2] + seglength] >= 0? 2 * m_perm[startpos[2] + seglength] - 1
                          : -2 * m_perm[startpos[2] + seglength]);             // f
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v1]) || (v0 == v1) || (v2 == v3))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[0];
                            op->m_edges[1] = startpos[2] - seglength;
                            seq.push_back(op);
                        }
                    }
                    else    // orientations negative
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[1] - seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[1] - 2 * seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength + 1;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[2] - 2 * seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        op = new Operation;
                        op->m_type = Operation::OPERATION_DUPLICATION;
                        op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                        op->m_edges[1] = op->m_edges[0] + seglength;
                        seq.push_back(op);
                        // if merge of two components or we created a selfloop
                        v0 = (m_perm[startpos[0] - seglength + 1] >= 0? 2 * m_perm[startpos[0] - seglength + 1] - 1
                          : -2 * m_perm[startpos[0] - seglength + 1]);     // last point of A
                        v1 = (m_perm[startpos[0] + 1] >= 0? 2 * m_perm[startpos[0] + 1] - 1
                          : -2 * m_perm[startpos[0] + 1]);                 // a
                        v2 = (m_perm[startpos[1] + 1] >= 0? 2 * m_perm[startpos[1] + 1] - 1
                          : -2 * m_perm[startpos[1] + 1]);                 // c
                        v3 = (m_perm[startpos[2] - seglength] >= 0? 2 * m_perm[startpos[2] - seglength]
                          : -2 * m_perm[startpos[2] - seglength] - 1);     // f
                        if ((m_bridgeCompID[v0] != m_bridgeCompID[v1]) || (v0 == v2) || (v1 == v3))
                        {
                            op = new Operation;
                            op->m_type = Operation::OPERATION_REVERSAL;
                            op->m_edges[0] = startpos[1] - seglength;
                            op->m_edges[1] = startpos[2] - 2 * seglength;
                            seq.push_back(op);
                        }
                    }
                } 
            }
            else if ((ori[0] == 1) && (ori[1] == 1) && (ori[2] == -1))
            {
                if (startpos[2] - startpos[1] + 1 < 2 * seglength)
                    seglength = (startpos[2] - startpos[1] + 1) / 2;
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[1];
                op->m_edges[1] = startpos[2] - seglength + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL; 
                op->m_edges[0] = startpos[0] + seglength;
                op->m_edges[1] = startpos[2] + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[0];
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[0];
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] + seglength] >= 0? 2 * m_perm[startpos[0] + seglength] - 1
                  : -2 * m_perm[startpos[0] + seglength]);      // b
                v1 = (m_perm[startpos[1] - 1] >= 0? 2 * m_perm[startpos[1] - 1]
                  : -2 * m_perm[startpos[1] - 1] - 1);          // c  
                v2 = (m_perm[startpos[2] + 1] >= 0? 2 * m_perm[startpos[2] + 1] - 1
                  : -2 * m_perm[startpos[2] + 1]);              // e
                v3 = (m_perm[startpos[2] - seglength] >= 0? 2 * m_perm[startpos[2] - seglength]
                  : -2 * m_perm[startpos[2] - seglength] - 1);  // f
                if (m_bridgeCompID[v0] != m_bridgeCompID[v1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength + 1;
                    op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                    seq.push_back(op);
                    if ((v0 == v3) && (startpos[2] - startpos[1] + 1 > 2 * seglength))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + seglength;
                        op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] - seglength + 1;
                        seq.push_back(op);
                    }
                    if (v1 == v2)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                    }
                }   
            }
            else if ((ori[0] == 1) && (ori[1] == -1) && (ori[2] == 1))
            {
                if (startpos[1] - startpos[0] + 1 < 2 * seglength)
                    seglength = (startpos[1] - startpos[0] + 1) / 2;
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0];
                op->m_edges[1] = startpos[1] - seglength + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[1] + 1;
                op->m_edges[1] = startpos[2] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[1] - 2 * seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[1] - 2 * seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] - 1] >= 0? 2 * m_perm[startpos[0] - 1] 
                  : -2 * m_perm[startpos[0] - 1] - 1);          // a
                v1 = (m_perm[startpos[1] + 1] >= 0? 2 * m_perm[startpos[1] + 1] - 1
                  : -2 * m_perm[startpos[1] + 1]);              // c  
                v2 = (m_perm[startpos[1] - seglength] >= 0? 2 * m_perm[startpos[1] - seglength]
                  : -2 * m_perm[startpos[1] - seglength] - 1);  // d
                v3 = (m_perm[startpos[2] + seglength] >= 0? 2 * m_perm[startpos[2] + seglength] - 1
                  : -2 * m_perm[startpos[2] + seglength]);      // f
                if (m_bridgeCompID[v0] != m_bridgeCompID[v2])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[0];
                    op->m_edges[1] = startpos[2] - seglength;
                    seq.push_back(op);
                    if ((v0 == v1) && (startpos[2] - startpos[1] > 1))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] - 1;
                        seq.push_back(op);
                    }
                    if ((v2 == v3) && (startpos[1] - startpos[0] + 1 > 2 * seglength))
                    { 
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] + seglength - 1;
                        op->m_edges[1] = startpos[2] - seglength;
                        seq.push_back(op);
                    }
                }
            }
            else if ((ori[0] == -1) && (ori[1] == 1) && (ori[2] == 1))
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0] + 1;
                op->m_edges[1] = startpos[1] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0] - seglength + 1;
                op->m_edges[1] = startpos[2];
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[2] - 2 * seglength;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op); 
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[2] - 2 * seglength;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op); 
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] + 1] >= 0? 2 * m_perm[startpos[0] + 1] - 1 
                  : -2 * m_perm[startpos[0] + 1]);              // a
                v1 = (m_perm[startpos[0] - seglength] >= 0? 2 * m_perm[startpos[0] - seglength] 
                  : -2 * m_perm[startpos[0] - seglength] - 1);  // b  
                v2 = (m_perm[startpos[1] + seglength] >= 0? 2 * m_perm[startpos[1] + seglength] - 1
                  : -2 * m_perm[startpos[1] + seglength]);      // d
                v3 = (m_perm[startpos[2] - 1] >= 0? 2 * m_perm[startpos[2] - 1]
                  : -2 * m_perm[startpos[2] - 1] - 1);          // e
                if (m_bridgeCompID[v0] != m_bridgeCompID[v1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[0] - seglength + 1;
                    op->m_edges[1] = op->m_edges[0] + startpos[2] - startpos[1] - seglength;
                    seq.push_back(op);
                    if ((v0 == v3) && (startpos[1] - startpos[0] > 1))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[2] - 2 * seglength - startpos[1] + startpos[0] + 1;
                        op->m_edges[1] = startpos[2] - 2 * seglength;
                        seq.push_back(op);
                    }
                    if (v1 == v2)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL; 
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength;
                        seq.push_back(op);
                    }
                }
            }
            else if ((ori[0] == 1) && (ori[1] == -1) && (ori[2] == -1))
            {
                if (startpos[1] - startpos[0] + 1 < 2 * seglength)
                    seglength = (startpos[1] - startpos[0] + 1) / 2;
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0] + seglength;
                op->m_edges[1] = startpos[1] + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0];
                op->m_edges[1] = startpos[2] - seglength + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[2] - 3 * seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] - 1] >= 0? 2 * m_perm[startpos[0] - 1] 
                  : -2 * m_perm[startpos[0] - 1] - 1);          // a
                v1 = (m_perm[startpos[0] + seglength] >= 0? 2 * m_perm[startpos[0] + seglength] - 1 
                  : -2 * m_perm[startpos[0] + seglength]);      // b  
                v2 = (m_perm[startpos[1] + 1] >= 0? 2 * m_perm[startpos[1] + 1] - 1
                  : -2 * m_perm[startpos[1] + 1]);              // c
                v3 = (m_perm[startpos[2] - seglength] >= 0? 2 * m_perm[startpos[2] - seglength]
                  : -2 * m_perm[startpos[2] - seglength] - 1);  // f
                if (m_bridgeCompID[v0] != m_bridgeCompID[v1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[0];
                    op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] - seglength;
                    seq.push_back(op);
                    if ((v1 == v3) && (startpos[1] - startpos[0] + 1 > 2 * seglength))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - seglength;
                        op->m_edges[1] = startpos[2] - 3 * seglength + 1;
                        seq.push_back(op);
                    }
                    if (v0 == v2)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0];
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                    }
                }
            }
            else if ((ori[0] == -1) && (ori[1] == 1) && (ori[2] == -1))
            {
                if (startpos[2] - startpos[1] + 1 < 2 * seglength)
                    seglength = (startpos[2] - startpos[1] + 1) / 2;
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;    
                op->m_edges[0] = startpos[0] - seglength + 1;
                op->m_edges[1] = startpos[1];
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[1] + seglength;
                op->m_edges[1] = startpos[2] + 1;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[1] - seglength;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[1] - seglength;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] - seglength] >= 0? 2 * m_perm[startpos[0] - seglength] 
                  : -2 * m_perm[startpos[0] - seglength] - 1);  // b
                v1 = (m_perm[startpos[1] - 1] >= 0? 2 * m_perm[startpos[1] - 1]
                  : -2 * m_perm[startpos[1] - 1] - 1);          // c  
                v2 = (m_perm[startpos[1] + seglength] >= 0? 2 * m_perm[startpos[1] + seglength] - 1
                  : -2 * m_perm[startpos[1] + seglength]);      // d
                v3 = (m_perm[startpos[2] + 1] >= 0? 2 * m_perm[startpos[2] + 1] - 1
                  : -2 * m_perm[startpos[2] + 1]);              // e
                if (m_bridgeCompID[v0] != m_bridgeCompID[v1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[0] - seglength + 1;
                    op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                    seq.push_back(op);
                    if ((v0 == v2) && (startpos[2] - startpos[1] + 1 > 2 * seglength))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[0] + startpos[2] - startpos[1] - 3 * seglength + 2;
                        seq.push_back(op);
                    }
                    if ((v1 == v3) && (startpos[1] - startpos[0] > 1))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + startpos[2] - startpos[1] - 2 * seglength + 2;
                        op->m_edges[1] = startpos[2] - 2 * seglength + 1;
                        seq.push_back(op);
                    }
                }
            }
            else if ((ori[0] == -1) && (ori[1] == -1) && (ori[2] == 1))
            {
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[1] - seglength + 1;
                op->m_edges[1] = startpos[2];
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_REVERSAL;
                op->m_edges[0] = startpos[0] + 1;
                op->m_edges[1] = startpos[2] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[0] - seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = startpos[0] - seglength + 1;
                op->m_edges[1] = op->m_edges[0] + seglength;
                seq.push_back(op);
                // if merge perform additional moves
                v0 = (m_perm[startpos[0] + 1] >= 0? 2 * m_perm[startpos[0] + 1] - 1 
                  : -2 * m_perm[startpos[0] + 1]);              // a
                v1 = (m_perm[startpos[1] - seglength] >= 0? 2 * m_perm[startpos[1] - seglength]
                  : -2 * m_perm[startpos[1] - seglength] - 1);  // d  
                v2 = (m_perm[startpos[2] - 1] >= 0? 2 * m_perm[startpos[2] - 1]
                  : -2 * m_perm[startpos[2] - 1] - 1);          // e
                v3 = (m_perm[startpos[2] + seglength] >= 0? 2 * m_perm[startpos[2] + seglength] - 1
                  : -2 * m_perm[startpos[2] + seglength] - 1);  // f
                if (m_bridgeCompID[v0] != m_bridgeCompID[v1])
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_REVERSAL;
                    op->m_edges[0] = startpos[2] - startpos[1] + startpos[0];
                    op->m_edges[1] = startpos[2] - seglength;
                    seq.push_back(op);
                    if ((v0 == v2) && (startpos[2] - startpos[0] > 1))
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] + 1;
                        op->m_edges[1] = startpos[0] + startpos[2] - startpos[1];
                        seq.push_back(op);
                    }
                    if (v1 == v3)
                    {
                        op = new Operation;
                        op->m_type = Operation::OPERATION_REVERSAL;
                        op->m_edges[0] = startpos[0] - seglength + 1;
                        op->m_edges[1] = startpos[2] - seglength;
                        seq.push_back(op);
                    }
                }   
            }
            else 
            {
                cerr << "something went wrong in findType3Sequences!\n";
                continue;
            }
            m_sequences.push_back(seq);
            
            // close the brackets of getting all triplets
                    } 
                }
            }
            // end closing brackets
            
        } // endif m_rddDegree[2*i] >= 4
    } // end for all positions

    return m_sequences.size() - oldSize;
}


int DcjdDist::findTranspositionDuplications()
{
    unsigned int oldSize = m_sequences.size();
    int dupllength;
    vector<Operation*> seq;         // a sequence of operations
    Operation* op;                  // an operation of the sequence
    
    for (int i = 0; i < m_permSize - 1; i++)
    {
        for (int j = i+1; j < m_permSize; j++)
        { 
            // check if seed
            if ((m_perm[i] == m_perm[j])        
              && ((i == 0) || (j == i+1) || (m_perm[i-1] != m_perm[j-1]))) 
            {
                dupllength = 1;
                // extend
                while ((i + dupllength < j) && (j + dupllength < m_permSize) 
                  && (m_perm[i+dupllength] == m_perm[j+dupllength]))
                    dupllength++;
                // remove second occurance    
                seq.clear();
                if (i + dupllength < j)       // real transposition duplication
                {
                    op = new Operation;
                    op->m_type = Operation::OPERATION_DELETION;
                    op->m_edges[0] = j + dupllength;
                    op->m_sequence.reserve(j - i - dupllength);                                         
                    for (int k = i + dupllength; k < j; k++)
                        op->m_sequence.push_back(m_perm[k]);
                    seq.push_back(op);
                }
                op = new Operation;
                op->m_type = Operation::OPERATION_DUPLICATION;
                op->m_edges[0] = i;
                op->m_edges[1] = j;
                seq.push_back(op);
                m_sequences.push_back(seq);
                // remove first occurance (if not a real transposition duplication, the effect would be the same)
                if (i + dupllength < j)       // real transposition duplication
                {
                    seq.clear();
                    op = new Operation;
                    op->m_type = Operation::OPERATION_DELETION;
                    op->m_edges[0] = i;
                    op->m_sequence.reserve(j - i - dupllength);                                         
                    for (int k = i + dupllength; k < j; k++)
                        op->m_sequence.push_back(m_perm[k]);
                    dupllength += op->m_sequence.size();
                    seq.push_back(op);
                    op = new Operation;
                    op->m_type = Operation::OPERATION_DUPLICATION;
                    op->m_edges[0] = i;
                    op->m_edges[1] = i + dupllength;
                    seq.push_back(op);
                    m_sequences.push_back(seq);
                }
            }
        }
    }    
    return m_sequences.size() - oldSize;    
}
 

int DcjdDist::findSpecialSequences()
{
    int oldSize = m_sequences.size();   // old number of sequences
    int v0, v1, v2, v3, v4;             // vertices in the RDD
    stack<int> expand;                  // vertices to expand (depth first search)
    int targetComp;                     // component ID of vertex a
    bool inc;                           // true if we have to search in increasing direction
    vector<Operation*> seq;             // a sequence of operations
    Operation* op;                      // an operation of the sequence
    int seglength;                      // length of an inserted segment
    
    // **** find deletion - reversal sequences *****
    for (int i = 0; i < m_permSize - 1; i++)
    {        
        if (m_perm[i] >= 0)
            v0 = 2 * m_perm[i];
        else
            v0 = -2 * m_perm[i] - 1;
        if (m_perm[i+1] >= 0)
            v1 = 2 * m_perm[i+1] - 1;
        else
            v1 = -2 * m_perm[i+1];
        if (!(((m_bridgeStatus[v0] == 1) && (m_bridges[v0][0].second == v1))   // not a 1-bridge
          || ((m_bridgeStatus[v1] == 1) && (m_bridges[v1][0].second == v0))))
            continue;
        // partition component - v1 side will get ID m_compID
        m_bridgeCompID[v1] = m_compID; 
        expand.push(v1);
        while (!expand.empty())
        {
            v2 = expand.top();
            expand.pop();        
            for (int j = 0; j < m_rddDegree[v2]; j++)
            {
                v3 = m_rddEdges[v2][j];
                if (v3 == v0)       // ignore the 1-bridge
                    continue;
                if (m_bridgeCompID[v3] == m_compID)    // already on stack
                    continue;
                m_bridgeCompID[v3] = m_compID;
                expand.push(v3);
            }
        }
        // search for partner edges
        for (int j = 0; j < m_permSize - 1; j++)
        {
            if (i == j)
                continue;
            if (m_perm[j] >= 0)
                v2 = 2 * m_perm[j];
            else
                v2 = -2 * m_perm[j] - 1;
            if (m_perm[j+1] >= 0)
                v3 = 2 * m_perm[j+1] - 1;
            else
                v3 = -2 * m_perm[j+1];
            if (m_bridgeCompID[v2] == m_bridgeCompID[v0])
            {
                if (v0 == v2)
                    continue;
                targetComp = m_compID;
                if (v3 & 1)
                {
                    v4 = v3 - 2;
                    inc = false;
                }
                else
                {
                    v4 = v3 + 2;
                    inc = true;    
                }
            }
            else if (m_bridgeCompID[v2] == m_bridgeCompID[v1])
            {
                if (v1 == v3)
                    continue;
                targetComp = m_bridgeCompID[v0];
                if (v2 & 1)
                {
                    v4 = v2 - 2;
                    inc = false;
                }
                else
                {
                    v4 = v2 + 2;
                    inc = true;
                }    
            }
            else              // different component
                continue;
            // search in steps of 2 for a fitting node
            if (inc)
            {
                while ((v4 < 2 * m_numMarkers - 3) && ((m_bridgeCompID[v4] != targetComp) || (v4 == v0) || (v4 == v1)))
                    v4 += 2;
            }
            else    
            {
                while ((v4 > 0) && ((m_bridgeCompID[v4] != targetComp) || (v4 == v0) || (v4 == v1)))
                    v4 -= 2;
            }
            if ((v4 <= 0) || (v4 >= 2 * m_numMarkers - 3))   // no fitting vertex found
                continue;
            if (((targetComp == m_compID) && (m_perm[j+1] - (inc? -v4 / 2 : (v4 + 1) / 2) > m_numMarkers / 3))
              || ((targetComp != m_compID) && ((inc? v4 / 2 : -(v4 + 1) / 2) - m_perm[j] > m_numMarkers / 3)))
                continue;
            // create the sequence of operations
            seq.clear();
            op = new Operation;
            op->m_type = Operation::OPERATION_DELETION;
            op->m_edges[0] = j + 1;
            if (targetComp == m_compID)                    // inserted segment must be adjacent to y
            {
                op->m_sequence.reserve(m_perm[j+1] - (inc? -v4 / 2 : (v4 + 1) / 2));
                for (int k = (inc? -v4 / 2 : (v4 + 1) / 2); k < m_perm[j+1]; k++)
                    op->m_sequence.push_back(k);
            }
            else                                           // inserted segment must be adjacent to x
            {
                op->m_sequence.reserve((inc? v4 / 2 : -(v4 + 1) / 2) - m_perm[j]);
                for (int k = m_perm[j] + 1; k <= (inc? v4 / 2 : -(v4 + 1) / 2); k++)
                    op->m_sequence.push_back(k);
            }
            seglength = op->m_sequence.size();
            seq.push_back(op);
            op = new Operation;
            op->m_type = Operation::OPERATION_REVERSAL;
            if (i < j)
            {
                op->m_edges[0] = i + 1;
                if (targetComp == m_compID)                // don't revert inserted segment
                    op->m_edges[1] = j + 1;
                else                                       // revert inserted segment
                    op->m_edges[1] = j + 1 +  seglength; 
            }
            else
            {
                if (targetComp == m_compID)                // revert inserted segment
                    op->m_edges[0] = j + 1;
                else                                       // don't revert inserted segment
                    op->m_edges[0] = j + 1 + seglength;
                op->m_edges[1] = i + 1 + seglength;
            }
            seq.push_back(op);
            if (((seq[0]->m_edges[0] == 34) && (seq[1]->m_edges[0] == 20) && (seq[1]->m_edges[1] == 34))
              || ((seq[0]->m_edges[0] == 130) && (seq[1]->m_edges[0] == 14) && (seq[1]->m_edges[1] == 193)))
            {
            
            m_sequences.push_back(seq);    
            }
            else
            {                         
                for (unsigned int deb = 0; deb < seq.size(); deb++)
                    delete seq[deb];
            }
            
        }
        // undo relabeling of comps
        for (int j = 0; j < 2 * m_numMarkers - 2; j++)
        {
            if (m_bridgeCompID[j] == m_compID) 
                m_bridgeCompID[j] = m_bridgeCompID[v0];
        }  
    }
    return m_sequences.size() - oldSize;
}
        

int DcjdDist::getScore(Operation& f_op, int f_lb, int* f_score)
{
    int newLb = 0;                    // the new lower bound
    int v0, v1, v2, v3, v4, v5, v6, v7;   // vertices in the RDD
    int* degrees = new int[m_numMarkers * 2 - 2];   // copy of m_rddDegrees
    int** edges = new int*[m_numMarkers * 2 - 2];   // copy of m_rddEdges
    
    // make a copy of the edges
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
        edges[i] = new int[m_permSize];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        degrees[i] = m_rddDegree[i];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        for (int j = 0; j < degrees[i]; j++)
            edges[i][j] = m_rddEdges[i][j];
    }
    // temporary apply the operation
    if (f_op.m_type == Operation::OPERATION_DUPLICATION)
    {
        // remove edges in the RDD
        for (int i = f_op.m_edges[0] + 1; i <= f_op.m_edges[1]; i++)
        {
            if (m_perm[i-1] < 0)
                v0 = 2 * (-m_perm[i-1]) - 1;
            else
                v0 = 2 * m_perm[i-1];
            if (m_perm[i] < 0)
                v1 = 2 * (-m_perm[i]);
            else
                v1 = 2 * m_perm[i] - 1;
            for (int j = 0; j < m_rddDegree[v0]; j++)
            {
                if (m_rddEdges[v0][j] == v1)
                {
                    m_rddEdges[v0][j] = m_rddEdges[v0][m_rddDegree[v0]-1];
                    m_rddDegree[v0]--;
                    break;
                }
            }
            for (int j = 0; j < m_rddDegree[v1]; j++)
            {
                if (m_rddEdges[v1][j] == v0)
                {
                    m_rddEdges[v1][j] = m_rddEdges[v1][m_rddDegree[v1]-1];
                    m_rddDegree[v1]--;
                    break;
                }
            }
        }
    }
    else if (f_op.m_type == Operation::OPERATION_REVERSAL)
    {
        // move edges in the RDD
        if (m_perm[f_op.m_edges[0]-1] < 0)
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        else
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        if (m_perm[f_op.m_edges[0]] < 0)
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        else
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        if (m_perm[f_op.m_edges[1]-1] < 0)
            v2 = 2 * (-m_perm[f_op.m_edges[1]-1]) - 1;
        else
            v2 = 2 * m_perm[f_op.m_edges[1]-1];
        if (m_perm[f_op.m_edges[1]] < 0)
            v3 = 2 * (-m_perm[f_op.m_edges[1]]);
        else
            v3 = 2 * m_perm[f_op.m_edges[1]] - 1;
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                m_rddEdges[v0][i] = v2;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v1]; i++)
        {
            if (m_rddEdges[v1][i] == v0)
            {
                m_rddEdges[v1][i] = v3;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v2]; i++)
        {
            if (m_rddEdges[v2][i] == v3)
            {
                m_rddEdges[v2][i] = v0;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v3]; i++)
        {
            if (m_rddEdges[v3][i] == v2)
            {
                m_rddEdges[v3][i] = v1;
                break;
            }
        }
    }
    else if (f_op.m_type == Operation::OPERATION_BLOCKINTERCHANGE)
    {
        // update the RDD
        if (m_perm[f_op.m_edges[0]-1] < 0)
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        else
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        if (m_perm[f_op.m_edges[0]] < 0)
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        else
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        if (m_perm[f_op.m_edges[1]-1] < 0)
            v2 = 2 * (-m_perm[f_op.m_edges[1]-1]) - 1;
        else
            v2 = 2 * m_perm[f_op.m_edges[1]-1];
        if (m_perm[f_op.m_edges[1]] < 0)
            v3 = 2 * (-m_perm[f_op.m_edges[1]]);
        else
            v3 = 2 * m_perm[f_op.m_edges[1]] - 1;
        if (m_perm[f_op.m_edges[2]-1] < 0)
            v4 = 2 * (-m_perm[f_op.m_edges[2]-1]) - 1;
        else
            v4 = 2 * m_perm[f_op.m_edges[2]-1];
        if (m_perm[f_op.m_edges[2]] < 0)
            v5 = 2 * (-m_perm[f_op.m_edges[2]]);
        else
            v5 = 2 * m_perm[f_op.m_edges[2]] - 1;
        if (m_perm[f_op.m_edges[3]-1] < 0)
            v6 = 2 * (-m_perm[f_op.m_edges[3]-1]) - 1;
        else
            v6 = 2 * m_perm[f_op.m_edges[3]-1];
        if (m_perm[f_op.m_edges[3]] < 0)
            v7 = 2 * (-m_perm[f_op.m_edges[3]]);
        else
            v7 = 2 * m_perm[f_op.m_edges[3]] - 1;
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                if (f_op.m_edges[2] != f_op.m_edges[3])
                    m_rddEdges[v0][i] = v5;
                else
                    m_rddEdges[v0][i] = v3;    
                break;
            }
        }
        if (f_op.m_edges[0] != f_op.m_edges[1])
        {
            for (int i = 0; i < m_rddDegree[v1]; i++)
            {
                if (m_rddEdges[v1][i] == v0)
                {
                    if (f_op.m_edges[1] != f_op.m_edges[2])
                        m_rddEdges[v1][i] = v4;
                    else
                        m_rddEdges[v1][i] = v6;
                    break;
                }
            }
        }
        if (f_op.m_edges[0] != f_op.m_edges[1])
        {
            for (int i = 0; i < m_rddDegree[v2]; i++)
            {
                if (m_rddEdges[v2][i] == v3)
                {
                    m_rddEdges[v2][i] = v7;
                    break;
                }
            }
        }
        if (f_op.m_edges[1] != f_op.m_edges[2])
        {
            for (int i = 0; i < m_rddDegree[v3]; i++)
            {
                if (m_rddEdges[v3][i] == v2)
                {
                    if (f_op.m_edges[2] != f_op.m_edges[3])
                        m_rddEdges[v3][i] = v6;
                    else
                        m_rddEdges[v3][i] = v0;
                    break;
                }
            }
        }
        if (f_op.m_edges[1] != f_op.m_edges[2])
        {
            for (int i = 0; i < m_rddDegree[v4]; i++)
            {
                if (m_rddEdges[v4][i] == v5)
                {
                    if (f_op.m_edges[0] != f_op.m_edges[1])
                        m_rddEdges[v4][i] = v1;
                    else
                        m_rddEdges[v4][i] = v7;
                    break;
                }
            }
        }
        if (f_op.m_edges[2] != f_op.m_edges[3])
        {
            for (int i = 0; i < m_rddDegree[v5]; i++)
            {
                if (m_rddEdges[v5][i] == v4)
                {
                    m_rddEdges[v5][i] = v0;
                    break;
                }
            }
        }
        if (f_op.m_edges[2] != f_op.m_edges[3])
        {
            for (int i = 0; i < m_rddDegree[v6]; i++)
            {
                if (m_rddEdges[v6][i] == v7)
                {
                    if (f_op.m_edges[1] != f_op.m_edges[2])
                        m_rddEdges[v6][i] = v3;
                    else
                        m_rddEdges[v6][i] = v1;
                    break;
                }
            }
        }
        for (int i = 0; i < m_rddDegree[v7]; i++)
        {
            if (m_rddEdges[v7][i] == v6)
            {
                if (f_op.m_edges[0] != f_op.m_edges[1])
                    m_rddEdges[v7][i] = v2;
                else
                    m_rddEdges[v7][i] = v4;
                break;
            }
        }
    }
    else if (f_op.m_type == Operation::OPERATION_DELETION)
    {
        // update the RDD
        if (m_perm[f_op.m_edges[0]-1] >= 0)
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        else
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        if (m_perm[f_op.m_edges[0]] >= 0)
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        else
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                m_rddEdges[v0][i] = m_rddEdges[v0][m_rddDegree[v0]-1];
                m_rddDegree[v0]--;
                break;
            }
        }     
        for (int i = 0; i < m_rddDegree[v1]; i++)
        {
            if (m_rddEdges[v1][i] == v0)
            {
                m_rddEdges[v1][i] = m_rddEdges[v1][m_rddDegree[v1]-1];
                m_rddDegree[v1]--;
                break;
            }
        }     
        if (f_op.m_sequence[0] >= 0)
            v3 = 2 * f_op.m_sequence[0] - 1;
        else 
            v3 = -2 * f_op.m_sequence[0];
        m_rddEdges[v0][m_rddDegree[v0]] = v3;
        m_rddDegree[v0]++;    
        m_rddEdges[v3][m_rddDegree[v3]] = v0;
        m_rddDegree[v3]++;
        for (unsigned int i = 1; i < f_op.m_sequence.size(); i++)
        {
            if (f_op.m_sequence[i-1] >= 0)
                v2 = 2 * f_op.m_sequence[i-1];
            else
                v2 = -2 * f_op.m_sequence[i-1] - 1;
            if (f_op.m_sequence[i] >= 0)
                v3 = 2 * f_op.m_sequence[i] - 1;
            else 
                v3 = -2 * f_op.m_sequence[i];
            m_rddEdges[v2][m_rddDegree[v2]] = v3;
            m_rddDegree[v2]++;    
            m_rddEdges[v3][m_rddDegree[v3]] = v2;
            m_rddDegree[v3]++;
        }
        if (f_op.m_sequence[f_op.m_sequence.size()-1] >= 0)
            v2 = 2 * f_op.m_sequence[f_op.m_sequence.size()-1];
        else
            v2 = -2 * f_op.m_sequence[f_op.m_sequence.size()-1] - 1;
        m_rddEdges[v2][m_rddDegree[v2]] = v1;
        m_rddDegree[v2]++;    
        m_rddEdges[v1][m_rddDegree[v1]] = v2;
        m_rddDegree[v1]++;
    }
    // calculate the new lower bound
    newLb = getLowerBound();
    // calculate the advanced score
    if (f_op.m_type == Operation::OPERATION_BLOCKINTERCHANGE)
    {
        f_score[0] = 30 * (f_lb - newLb);
        f_score[1] = -1;
    }
    else 
    {
        f_score[0] = 60 * (f_lb - newLb);
        f_score[1] = -1;
    }
    
    if (f_op.m_type == Operation::OPERATION_DELETION)
        f_score[1] = -2;
    
    f_score[2] = 0;
    f_score[3] = -m_numMarkers + 1;
    for (int i = 0; i < m_numMarkers - 1; i++)
    {
        if (m_rddDegree[2*i] == 1)       // missing element
            f_score[2] -= 2;
        else
            f_score[2] -= 2 * (m_rddDegree[2*i] - 1);
        for (int j = 0; j < m_rddDegree[2*i]; j++)
        {
            if (m_rddEdges[2*i][j] == 2*i+1)
                f_score[3]++;
        }    
    }
    
    f_score[2] += f_score[3];
    f_score[3] = 0;
    
    // undo the operation
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        m_rddDegree[i] = degrees[i];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        for (int j = 0; j < degrees[i]; j++)
            m_rddEdges[i][j] = edges[i][j];
        delete[] edges[i];
    }
    delete[] edges;
    delete[] degrees;
    // return the result
    if (f_op.m_type == Operation::OPERATION_BLOCKINTERCHANGE)
        return (f_lb - newLb);
    else if (f_op.m_type == Operation::OPERATION_DELETION)
        return 2 * (f_lb - newLb) - 1;
    else        
        return 2 * (f_lb - newLb);
}


void DcjdDist::getSequenceScore(std::vector<Operation*>& f_seq, int f_lb, int* f_score)
{
    int* degrees = new int[m_numMarkers * 2 - 2];   // copy of m_rddDegrees
    int** edges = new int*[m_numMarkers * 2 - 2];   // copy of m_rddEdges
    int* freq;                            // element frequencies
    int* perm = new int[m_permSize];
    int permSize = m_permSize;
    int* selfloops = new int[m_numMarkers * 2 - 2];

    // init the score
    f_score[0] = 0;    
    f_score[1] = 0;    
    f_score[2] = 0;    
    f_score[3] = 0;    
    // make a copy of the edges
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
        edges[i] = new int[m_permSize];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        degrees[i] = m_rddDegree[i];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        for (int j = 0; j < degrees[i]; j++)
            edges[i][j] = m_rddEdges[i][j];
    }
    // make a copy of the permutation
    for (int i = 0; i < m_permSize; i++)
        perm[i] = m_perm[i];
    // make a copy of the selfloops    
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
        selfloops[i] = m_numSelfloops[i];
    // temporary apply the sequence
    for (unsigned int opi = 0; opi < f_seq.size(); opi++)
    {
        performOperation(*(f_seq[opi]));
        if (f_seq[opi]->m_type == Operation::OPERATION_DUPLICATION)
            f_score[1] -= 1;
        else if (f_seq[opi]->m_type == Operation::OPERATION_REVERSAL)
            f_score[1] -= 1;
        else if (f_seq[opi]->m_type == Operation::OPERATION_BLOCKINTERCHANGE)
            f_score[1] -= 2;
        else if (f_seq[opi]->m_type == Operation::OPERATION_DELETION)
            f_score[1] -= 2;
    }
    // calculate the new score (f_score[1] already done)
    f_score[0] = 60 * (f_lb - getLowerBound()) / (-f_score[1]);
    
    freq = new int[m_numMarkers];
    for (int i = 0; i < m_numMarkers; i++)
        freq[i] = 0;
    for (int i = 0; i < m_permSize; i++)
        freq[abs(m_perm[i])]++;
    
    for (int i = 0; i < m_numMarkers; i++)
    {
        if (freq[i] == 0)
            f_score[2] -= 2;
        else
            f_score[2] -= 2 * (freq[i] - 1);
    }
    delete[] freq;
    for (int i = 1; i < m_permSize; i++)
    {
        if (m_perm[i] == m_perm[i-1] + 1)
            f_score[2]++;
    }
    // undo the operation
    for (int i = 0; i < permSize; i++)
        m_perm[i] = perm[i];
    m_permSize = permSize;
    delete[] perm;    
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)   
        m_numSelfloops[i] = selfloops[i];
    delete[] selfloops;    
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
        m_rddDegree[i] = degrees[i];
    for (int i = 0; i < m_numMarkers * 2 - 2; i++)
    {
        for (int j = 0; j < degrees[i]; j++)
            m_rddEdges[i][j] = edges[i][j];
        delete[] edges[i];
    }
    delete[] edges;
    delete[] degrees;
/*    
    cout << "calculated score: (" << f_score[0] << " | " << f_score[1] 
      << " | " << f_score[2] << " | " << f_score[3] << ")\n";
    cout << "*****\n";
    */  
}


int DcjdDist::compSeqScores(int* f_a, int* f_b)
{
    for (int i = 0; i < 4; i++)
    {
        if (f_a[i] < f_b[i])
            return -1;
        else if (f_a[i] > f_b[i])
            return 1;
    }
    return 0;
}


void DcjdDist::performOperation(Operation& f_op)
{
    int v0, v1, v2, v3, v4, v5, v6, v7;       // vertices in the RDD
    int pos;        // position in the permutation
    int segment;    // the current segment of the permutation
    
    if (f_op.m_type == Operation::OPERATION_INVALID)
        return;
    if (f_op.m_type == Operation::OPERATION_DUPLICATION)
    {
        // consistency check
        for (int i = f_op.m_edges[0], j = f_op.m_edges[1]; i < f_op.m_edges[1]; i++, j++)
        {
            if ((j >= m_permSize) || (m_perm[i] != m_perm[j]))
            {
                cerr << "invalid duplication!\n";
                exit(0);
            }
        }
        // remove edges in the RDD
        for (int i = f_op.m_edges[0] + 1; i <= f_op.m_edges[1]; i++)
        {
            if (m_perm[i-1] < 0)
                v0 = 2 * (-m_perm[i-1]) - 1;
            else
                v0 = 2 * m_perm[i-1];
            if (m_perm[i] < 0)
                v1 = 2 * (-m_perm[i]);
            else
                v1 = 2 * m_perm[i] - 1;
            for (int j = 0; j < m_rddDegree[v0]; j++)
            {
                if (m_rddEdges[v0][j] == v1)
                {
                    m_rddEdges[v0][j] = m_rddEdges[v0][m_rddDegree[v0]-1];
                    m_rddDegree[v0]--;
                    break;
                }
            }
            for (int j = 0; j < m_rddDegree[v1]; j++)
            {
                if (m_rddEdges[v1][j] == v0)
                {
                    m_rddEdges[v1][j] = m_rddEdges[v1][m_rddDegree[v1]-1];
                    m_rddDegree[v1]--;
                    break;
                }
            }
            if (v0 == v1)
                m_numSelfloops[v0]--;
        }
        // update the permutation
        for (int i = 0; i < m_permSize - f_op.m_edges[1]; i++)
            m_perm[f_op.m_edges[0] + i] = m_perm[f_op.m_edges[1] + i];
        m_permSize -= (f_op.m_edges[1] - f_op.m_edges[0]);    
        return;
    }
    if (f_op.m_type == Operation::OPERATION_REVERSAL)
    {
        // move edges in the RDD
        if (m_perm[f_op.m_edges[0]-1] < 0)
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        else
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        if (m_perm[f_op.m_edges[0]] < 0)
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        else
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        if (m_perm[f_op.m_edges[1]-1] < 0)
            v2 = 2 * (-m_perm[f_op.m_edges[1]-1]) - 1;
        else
            v2 = 2 * m_perm[f_op.m_edges[1]-1];
        if (m_perm[f_op.m_edges[1]] < 0)
            v3 = 2 * (-m_perm[f_op.m_edges[1]]);
        else
            v3 = 2 * m_perm[f_op.m_edges[1]] - 1;
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                m_rddEdges[v0][i] = v2;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v1]; i++)
        {
            if (m_rddEdges[v1][i] == v0)
            {
                m_rddEdges[v1][i] = v3;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v2]; i++)
        {
            if (m_rddEdges[v2][i] == v3)
            {
                m_rddEdges[v2][i] = v0;
                break;
            }
        }
        for (int i = 0; i < m_rddDegree[v3]; i++)
        {
            if (m_rddEdges[v3][i] == v2)
            {
                m_rddEdges[v3][i] = v1;
                break;
            }
        }
        if (v0 == v1)
            m_numSelfloops[v0]--;
        if (v2 == v3)    
            m_numSelfloops[v2]--;
        if (v0 == v2)    
            m_numSelfloops[v0]++;
        if (v1 == v3)    
            m_numSelfloops[v1]++;
        // update the permutation
        for (int i = f_op.m_edges[0]; i < f_op.m_edges[1]; i++)
            m_tmpPerm[i] = m_perm[i];
        for (int i = 0; i < f_op.m_edges[1] - f_op.m_edges[0]; i++)
            m_perm[f_op.m_edges[0] + i] = -m_tmpPerm[f_op.m_edges[1] - i - 1];
        return;
    }
    if (f_op.m_type == Operation::OPERATION_DCJCIRC)  // should never happen
        return;
    if (f_op.m_type == Operation::OPERATION_BLOCKINTERCHANGE)
    {
        // update the RDD
        if (m_perm[f_op.m_edges[0]-1] < 0)
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        else
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        if (m_perm[f_op.m_edges[0]] < 0)
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        else
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        if (m_perm[f_op.m_edges[1]-1] < 0)
            v2 = 2 * (-m_perm[f_op.m_edges[1]-1]) - 1;
        else
            v2 = 2 * m_perm[f_op.m_edges[1]-1];
        if (m_perm[f_op.m_edges[1]] < 0)
            v3 = 2 * (-m_perm[f_op.m_edges[1]]);
        else
            v3 = 2 * m_perm[f_op.m_edges[1]] - 1;
        if (m_perm[f_op.m_edges[2]-1] < 0)
            v4 = 2 * (-m_perm[f_op.m_edges[2]-1]) - 1;
        else
            v4 = 2 * m_perm[f_op.m_edges[2]-1];
        if (m_perm[f_op.m_edges[2]] < 0)
            v5 = 2 * (-m_perm[f_op.m_edges[2]]);
        else
            v5 = 2 * m_perm[f_op.m_edges[2]] - 1;
        if (m_perm[f_op.m_edges[3]-1] < 0)
            v6 = 2 * (-m_perm[f_op.m_edges[3]-1]) - 1;
        else
            v6 = 2 * m_perm[f_op.m_edges[3]-1];
        if (m_perm[f_op.m_edges[3]] < 0)
            v7 = 2 * (-m_perm[f_op.m_edges[3]]);
        else
            v7 = 2 * m_perm[f_op.m_edges[3]] - 1;
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                if (f_op.m_edges[2] != f_op.m_edges[3])
                    m_rddEdges[v0][i] = v5;
                else
                    m_rddEdges[v0][i] = v3;    
                break;
            }
        }
        if (f_op.m_edges[0] != f_op.m_edges[1])
        {
            for (int i = 0; i < m_rddDegree[v1]; i++)
            {
                if (m_rddEdges[v1][i] == v0)
                {
                    if (f_op.m_edges[1] != f_op.m_edges[2])
                        m_rddEdges[v1][i] = v4;
                    else
                        m_rddEdges[v1][i] = v6;
                    break;
                }
            }
        }
        if (f_op.m_edges[0] != f_op.m_edges[1])
        {
            for (int i = 0; i < m_rddDegree[v2]; i++)
            {
                if (m_rddEdges[v2][i] == v3)
                {
                    m_rddEdges[v2][i] = v7;
                    break;
                }
            }
        }
        if (f_op.m_edges[1] != f_op.m_edges[2])
        {
            for (int i = 0; i < m_rddDegree[v3]; i++)
            {
                if (m_rddEdges[v3][i] == v2)
                {
                    if (f_op.m_edges[2] != f_op.m_edges[3])
                        m_rddEdges[v3][i] = v6;
                    else
                        m_rddEdges[v3][i] = v0;
                    break;
                }
            }
        }
        if (f_op.m_edges[1] != f_op.m_edges[2])
        {
            for (int i = 0; i < m_rddDegree[v4]; i++)
            {
                if (m_rddEdges[v4][i] == v5)
                {
                    if (f_op.m_edges[0] != f_op.m_edges[1])
                        m_rddEdges[v4][i] = v1;
                    else
                        m_rddEdges[v4][i] = v7;
                    break;
                }
            }
        }
        if (f_op.m_edges[2] != f_op.m_edges[3])
        {
            for (int i = 0; i < m_rddDegree[v5]; i++)
            {
                if (m_rddEdges[v5][i] == v4)
                {
                    m_rddEdges[v5][i] = v0;
                    break;
                }
            }
        }
        if (f_op.m_edges[2] != f_op.m_edges[3])
        {
            for (int i = 0; i < m_rddDegree[v6]; i++)
            {
                if (m_rddEdges[v6][i] == v7)
                {
                    if (f_op.m_edges[1] != f_op.m_edges[2])
                        m_rddEdges[v6][i] = v3;
                    else
                        m_rddEdges[v6][i] = v1;
                    break;
                }
            }
        }
        for (int i = 0; i < m_rddDegree[v7]; i++)
        {
            if (m_rddEdges[v7][i] == v6)
            {
                if (f_op.m_edges[0] != f_op.m_edges[1])
                    m_rddEdges[v7][i] = v2;
                else
                    m_rddEdges[v7][i] = v4;
                break;
            }
        }
        if (f_op.m_edges[0] == f_op.m_edges[1])
        {
            if (v0 == v1)    
                m_numSelfloops[v0]--;
            if (v4 == v5)   
                m_numSelfloops[v4]--;
            if (v6 == v7)    
                m_numSelfloops[v6]--;
            if (v0 == v5)    
                m_numSelfloops[v0]++;
            if (v1 == v6)   
                m_numSelfloops[v1]++;
            if (v4 == v7)    
                m_numSelfloops[v4]++;
        }
        else if (f_op.m_edges[1] == f_op.m_edges[2])
        {
            if (v0 == v1)    
                m_numSelfloops[v0]--;
            if (v2 == v3)    
                m_numSelfloops[v2]--;
            if (v6 == v7)    
                m_numSelfloops[v6]--;
            if (v0 == v5)    
                m_numSelfloops[v0]++;
            if (v2 == v7)    
                m_numSelfloops[v2]++;
            if (v1 == v6)
                m_numSelfloops[v1]++;    
        }
        else if (f_op.m_edges[2] == f_op.m_edges[3])
        {
            if (v0 == v1)    
                m_numSelfloops[v0]--;
            if (v2 == v3)    
                m_numSelfloops[v2]--;
            if (v4 == v5)   
                m_numSelfloops[v4]--;
            if (v0 == v3)    
                m_numSelfloops[v0]++;
            if (v1 == v4)   
                m_numSelfloops[v1]++;
            if (v2 == v7)    
                m_numSelfloops[v2]++;
        }
        else
        {
            if (v0 == v1)    
                m_numSelfloops[v0]--;
            if (v2 == v3)    
                m_numSelfloops[v2]--;
            if (v4 == v5)   
                m_numSelfloops[v4]--;
            if (v6 == v7)    
                m_numSelfloops[v6]--;
            if (v0 == v5)    
                m_numSelfloops[v0]++;
            if (v1 == v4)   
                m_numSelfloops[v1]++;
            if (v2 == v7)    
                m_numSelfloops[v2]++;
            if (v3 == v6)
                m_numSelfloops[v3]++;
        } 
        // update the permutation
        for (int i = f_op.m_edges[0]; i < f_op.m_edges[3]; i++)
            m_tmpPerm[i] = m_perm[i];
        pos = f_op.m_edges[2];         
        segment = 3;
        for (int i = f_op.m_edges[0]; i < f_op.m_edges[3]; i++, pos++)
        {
            if ((pos == f_op.m_edges[3]) && (segment == 3))
            {
                pos = f_op.m_edges[1];
                segment = 2;
            }
            if ((pos == f_op.m_edges[2]) && (segment == 2))
            {
                pos = f_op.m_edges[0];
                segment = 1;
            }
            m_perm[i] = m_tmpPerm[pos];
        }        
        return;
    }
    if (f_op.m_type == Operation::OPERATION_DELETION)
    {
        if (f_op.m_sequence.size() == 0)    
            return;
        // update the RDD
        if (m_perm[f_op.m_edges[0]-1] >= 0)
            v0 = 2 * m_perm[f_op.m_edges[0]-1];
        else
            v0 = 2 * (-m_perm[f_op.m_edges[0]-1]) - 1;
        if (m_perm[f_op.m_edges[0]] >= 0)
            v1 = 2 * m_perm[f_op.m_edges[0]] - 1;
        else
            v1 = 2 * (-m_perm[f_op.m_edges[0]]);
        for (int i = 0; i < m_rddDegree[v0]; i++)
        {
            if (m_rddEdges[v0][i] == v1)
            {
                m_rddEdges[v0][i] = m_rddEdges[v0][m_rddDegree[v0]-1];
                m_rddDegree[v0]--;
                break;
            }
        }     
        for (int i = 0; i < m_rddDegree[v1]; i++)
        {
            if (m_rddEdges[v1][i] == v0)
            {
                m_rddEdges[v1][i] = m_rddEdges[v1][m_rddDegree[v1]-1];
                m_rddDegree[v1]--;
                break;
            }
        }     
        if (v0 == v1)
            m_numSelfloops[v0]--;
        if (f_op.m_sequence[0] >= 0)
            v3 = 2 * f_op.m_sequence[0] - 1;
        else 
            v3 = -2 * f_op.m_sequence[0];
        m_rddEdges[v0][m_rddDegree[v0]] = v3;
        m_rddDegree[v0]++;    
        m_rddEdges[v3][m_rddDegree[v3]] = v0;
        m_rddDegree[v3]++;
        if (v0 == v3)
            m_numSelfloops[v0]++;
        for (unsigned int i = 1; i < f_op.m_sequence.size(); i++)
        {
            if (f_op.m_sequence[i-1] >= 0)
                v2 = 2 * f_op.m_sequence[i-1];
            else
                v2 = -2 * f_op.m_sequence[i-1] - 1;
            if (f_op.m_sequence[i] >= 0)
                v3 = 2 * f_op.m_sequence[i] - 1;
            else 
                v3 = -2 * f_op.m_sequence[i];
            m_rddEdges[v2][m_rddDegree[v2]] = v3;
            m_rddDegree[v2]++;    
            m_rddEdges[v3][m_rddDegree[v3]] = v2;
            m_rddDegree[v3]++;
            if (v2 == v3)
                m_numSelfloops[v2]++;
        }
        if (f_op.m_sequence[f_op.m_sequence.size()-1] >= 0)
            v2 = 2 * f_op.m_sequence[f_op.m_sequence.size()-1];
        else
            v2 = -2 * f_op.m_sequence[f_op.m_sequence.size()-1] - 1;
        m_rddEdges[v2][m_rddDegree[v2]] = v1;
        m_rddDegree[v2]++;    
        m_rddEdges[v1][m_rddDegree[v1]] = v2;
        m_rddDegree[v1]++;
        if (v2 == v1)
            m_numSelfloops[v2]++;
        // update the permutation
        for (int i = f_op.m_edges[0]; i < m_permSize; i++)
            m_tmpPerm[i] = m_perm[i];
        for (unsigned int i = 0; i < f_op.m_sequence.size(); i++)
            m_perm[f_op.m_edges[0] + i] = f_op.m_sequence[i];
        for (int i = f_op.m_edges[0]; i < m_permSize; i++)
            m_perm[i + f_op.m_sequence.size()] = m_tmpPerm[i];
        m_permSize += f_op.m_sequence.size();    
        return;    
    }
}


int DcjdDist::getLowerBound()
{
    bool* selfloops = new bool[2 * m_numMarkers - 2];
    bool* visited = new bool[2 * m_numMarkers - 2];
    int slcount;
    int result = m_numMarkers - 1;
    stack<int> extend;
    int node;
    
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
        selfloops[i] = false;
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
        visited[i] = false;
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
    {
        if (visited[i])
            continue;
        slcount = 0;
        extend.push(i);
        while (!extend.empty())
        {
            node = extend.top();
            extend.pop();
            visited[node] = true;
            for (int j = 0; j < m_rddDegree[node]; j++)
            {
                if ((m_rddEdges[node][j] == node) && (!selfloops[node]))
                {
                    selfloops[node] = true;
                    slcount++;
                }
                else if (!visited[m_rddEdges[node][j]])
                    extend.push(m_rddEdges[node][j]);
            }
        }
        result--;    // new component
        result += (slcount / 2 + (slcount & 1));   // selfloops        
    }
    delete[] selfloops;
    delete[] visited;
    return result;
} 


/***************************************************************************
 * debug methods
 ***************************************************************************/

bool DcjdDist::checkRDD()
{
    int** graph = new int*[2 * m_numMarkers - 2];
    int* selfloops = new int[2 * m_numMarkers - 2];
    int v0, v1;
    
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
    {
        graph[i] = new int[2 * m_numMarkers - 2];
        for (int j = 0; j < 2 * m_numMarkers - 2; j++)
            graph[i][j] = 0;
        if (i & 1)
            graph[i][i-1] = 1;
        else
            graph[i][i+1] = 1;
    }
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
        selfloops[i] = 0;    
    for (int i = 0; i < m_permSize - 1; i++)
    {
        if (m_perm[i] >= 0)
            v0 = 2 * m_perm[i];
        else
            v0 = -2 * m_perm[i] - 1;
        if (m_perm[i+1] >= 0)
            v1 = 2 * m_perm[i+1] - 1;
        else
            v1 = -2 * m_perm[i+1];
        graph[v0][v1]++;
        graph[v1][v0]++;
        if (v0 == v1)
            selfloops[v0]++;
    }
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
    {
        for (int j = 0; j < m_rddDegree[i]; j++)
            graph[i][m_rddEdges[i][j]]--;
    }
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
    {
        for (int j = 0; j < 2 * m_numMarkers - 2; j++)
        {
            if (graph[i][j] != 0)
            {
                cout << "inconsitency in RDD (edge " << i << " " << j << ")\n";
                for (int i = 0; i < 2 * m_numMarkers - 2; i++)
                    delete[] graph[i];
                delete[] graph;    
                delete[] selfloops;
                return false;
            }
        }
    }    
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
        delete[] graph[i];
    delete[] graph;
    for (int i = 0; i < 2 * m_numMarkers - 2; i++)
    {
        if (selfloops[i] != m_numSelfloops[i])
        {
            cout << "inconsistency in Selfloops (vertex " << i << ", " 
              << selfloops[i] << " vs " << m_numSelfloops[i] << ")\n";
            delete[] selfloops;
            return false;
        }
    }
    delete[] selfloops;
    return true;
}
 

void DcjdDist::printOperation(Operation f_op)
{
    if (f_op.m_type == Operation::OPERATION_INVALID)
        cout << "Inv\n";
    else if (f_op.m_type == Operation::OPERATION_DUPLICATION)
        cout << "Dup(" << f_op.m_edges[0] << ", " << f_op.m_edges[1] << ")\n";        
    else if (f_op.m_type == Operation::OPERATION_REVERSAL)
        cout << "Rev(" << f_op.m_edges[0] << ", " << f_op.m_edges[1] << ")\n";        
    else if (f_op.m_type == Operation::OPERATION_DCJCIRC)
        cout << "Dcj(" << f_op.m_edges[0] << ", " << f_op.m_edges[1] << ")\n";        
    else if (f_op.m_type == Operation::OPERATION_BLOCKINTERCHANGE)
        cout << "BI(" << f_op.m_edges[0] << ", " << f_op.m_edges[1] << ", " 
          << f_op.m_edges[2] << ", " << f_op.m_edges[3] << ")\n";
    else if (f_op.m_type == Operation::OPERATION_DELETION)
    {
        cout << "Del(" << f_op.m_edges[0] << ",";
        for (unsigned int i = 0; i < f_op.m_sequence.size(); i++)
            cout << " " << f_op.m_sequence[i];
        cout << ")\n";
    } 
}


/***************************************************************************/

} // end namespace dcjdDist
