/**
 * filename: start.cpp
 * author: Martin Bader
 * begin: 20.12.2007
 * last change: 22.03.2010
 *
 * Start file. Small UI.
 * 
 * Copyright (C) 2007-2010  Martin Bader
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
  
#include "dcjdDist.h"

#include <fstream>
#include <iostream>
  
using namespace std;


/** 
 * Reads a permutation from a file. This is not yet the final format.
 * @param f_filename  name of the file
 * @param f_numMarkers  the number of different markers will be written here
 * @param f_permSize  the permutation size will be written here
 * @param f_perm  reference of the permutation target pointer
 */
void readFile(const char* f_filename, int& f_numMarkers, int& f_permSize, int*& f_perm)
{
    ifstream input;    // input file
    
    // read file
    input.open(f_filename);
    input >> f_permSize;
    input >> f_numMarkers;
    f_permSize += 2;             // augmented perm
    f_numMarkers += 2;
    f_perm = new int[10 * f_permSize];
    f_perm[0] = 0;
    f_perm[f_permSize-1] = f_numMarkers - 1;
    for (int i = 1; i < f_permSize - 1; i++)
        input >> f_perm[i];
    input.close();
}


/**
 * Main function.
 * Usage not yet specified, as it is just a collection of experiments.
 * @param f_argc  number of arguments
 * @param f_argv  the arguments
 * @return  program exit code
 */
int main(int f_argc, char** f_argv)
{    
    dcjdDist::DcjdDist dcjdd;      // the program to calculate the distance
    int numMarkers;                // number of markers in the genome
    int permSize;                  // permutation size
    int* perm;                   // the permutation to sort
    
    cout << "*****************************************************************\n";
    cout << "dcjdDist v0.1.1  Copyright (C) 2009-2010  Martin Bader\n";
    cout << "This program comes with ABSOLUTELY NO WARRANTY.\n";
    cout << "This is free software, and you are welcome to redistribute it\n";
    cout << "under certain conditions.\n";
    cout << "For details about the licence, read the file gpl.txt\n";
    cout << "*****************************************************************\n";
    if (f_argc != 2)
    {
        cout << "Usage: dcjddist infile\n";
        return 0;
    }
    readFile(f_argv[1], numMarkers, permSize, perm);
    dcjdd.sort(numMarkers, permSize, perm);
    // clean up
    delete[] perm;
    return 0;    
}
